package.preload['minidriver'] = (function (...)
--Copyright 2018 Control4 Corporation. All rights reserved.

--[[
	In general, you should merge
]]

function ON_DRIVER_INIT.minidriver_init()
	PASSTHROUGH_PROXY = 5001		-- set this to the proxy ID that should handle all passthrough commands from minidrivers
	SWITCHER_PROXY = 5001					-- set this to the proxy ID of the SET_INPUT capable device that has the RF_MINI_APP connections (may be the same as PASSTHROUGH_PROXY)
	USES_DEDICATED_SWITCHER = false			-- set this to false if the driver did not need the dedicated avswitch proxy (e.g. this is a TV/receiver)
	MINIAPP_BINDING_START = 3051				-- set this to the first binding ID in the XML for the RF_MINI_APP connections
	MINIAPP_BINDING_END = 3075				-- set this to the last binding ID in the XML for the RF_MINI_APP connections
	MINIAPP_TYPE = 'UM_LG_2019'				-- set this to your unique name as defined in the minidriver SERVICE_IDS table
end

function ON_DRIVER_LATEINIT.minidriver_init()
	if (USES_DEDICATED_SWITCHER) then
		HideProxyInAllRooms (SWITCHER_PROXY)
	end
	RegisterRooms ()
end

function OnSystemEvent (event)
	local eventname = string.match (event, '.-name="(.-)"')
	if (eventname == 'OnPIP') then
		RegisterRooms ()
	end
end

function OnWatchedVariableChanged (idDevice, idVariable, strValue)
	if (RoomIDs and RoomIDs [idDevice]) then
		local roomId = tonumber (idDevice)
		if (idVariable == 1000) then
			local deviceId = tonumber (strValue) or 0
			RoomIDSources [roomId] = deviceId
		end
	end
end

--[[
function ReceivedFromProxy (idBinding, strCommand, tParams)
	if (idBinding == nil or strCommand == nil) then
		return	-- this shouldn't happen, but is a good defensive programming practice
	end

	if (tParams == nil) then
		tParams = {}	-- this often happens, especially with basic transport commands.  This protects against nil table referencing.
	end

	if (idBinding == SWITCHER_PROXY and strCommand == 'PASSTHROUGH') then
		idBinding = PASSTHROUGH_PROXY
		strCommand = tParams.PASSTHROUGH_COMMAND
	end

	if (idBinding == SWITCHER_PROXY) then
		if (strCommand == 'SET_INPUT') then
			local input = tonumber (tParams.INPUT)
			if (input >= MINIAPP_BINDING_START and input <= MINIAPP_BINDING_END) then
				-- Get the device ID of the proxy handling the miniapp switch on this driver
				local proxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

				-- Get the device ID of the minidriver proxy connected to the requested input on this driver
				local appProxyId = C4:GetBoundProviderDevice (proxyDeviceId, input)

				-- Get the device ID of the minidriver protocol connected to the minidriver proxy
				local appDeviceId = C4:GetBoundProviderDevice (appProxyId, 5001)

				-- get the details for the app for this kind of universal-minidriver-compatible type
				local appId = GetRelevantUniversalAppId (appDeviceId, MINIAPP_TYPE)
				local appName = GetRelevantUniversalAppId (appDeviceId, 'APP_NAME')

				-- there is now enough information to launch the application using the protocol for this device

				-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
				-- =-=-=-=-=-=-=-= YOUR DEVICE-SPECIFIC APP LAUNCHING CODE GOES HERE... -=-=-=-=-=-=-=
				-- =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

				-- there is now enough information to launch the application using the protocol for this device

				if ((Properties ['Passthrough Mode'] or 'On') ~= 'On') then
					local passthroughProxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), PASSTHROUGH_PROXY))
					local _timer = function (timer)
						print ('Looking for ' .. appProxyId)
						for roomId, deviceId in pairs (RoomIDSources) do
							if (deviceId == appProxyId) then
								C4:SendToDevice (roomId, 'SELECT_VIDEO_DEVICE', {deviceid = passthroughProxyDeviceId})
							end
						end
					end
					C4:SetTimer (500, _timer)
				end
			end
		end
	end
end
--]]

function GetRelevantUniversalAppId (deviceId, source)
	local vars = C4:GetDeviceVariables (deviceId)
	for _, var in pairs (vars) do
		if (var.name == source) then
			return (var.value)
		end
	end
	if (source ~= 'APP_ID') then
		-- try getting pre-universal minidriver app ID to launch.
		return (GetRelevantUniversalAppId (deviceId, 'APP_ID'))
	end
end

function HideProxyInAllRooms (idBinding)
	idBinding = idBinding or 0
	if (idBinding == 0) then return end -- silently fail if no binding passed in

	-- Get Bound Proxy's Device ID / Name.
	local id, name = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

	dbg ('Hiding ' .. name .. ' in all rooms')

	-- Send hide command to all rooms, for 'ALL' Navigator groups.
	for roomId, roomName in pairs (C4:GetDevicesByC4iName ('roomdevice.c4i') or {}) do
		dbg ('Hiding ' .. name .. ' in ' .. roomName)
		C4:SendToDevice (roomId, 'SET_DEVICE_HIDDEN_STATE', {PROXY_GROUP = 'ALL', DEVICE_ID = id, IS_HIDDEN = true})
	end
end

function RegisterRooms ()
	RoomIDs = C4:GetDevicesByC4iName ('roomdevice.c4i')
	RoomIDSources = {}
	for roomId, _ in pairs (RoomIDs) do
		RoomIDSources [roomId] = tonumber (C4:GetDeviceVariable (roomId, 1000)) or 0
		C4:UnregisterVariableListener (roomId, 1000)
		C4:RegisterVariableListener (roomId, 1000)
	end
end
 end)
package.preload['projector.projector_proxy_notifies'] = (function (...)
--[[=============================================================================
    Projector Proxy Notification Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--[[
	Notify: INPUT_CHANGED
	Sent to the proxy to indicate the currently selected Input has changed
	Parameters
		bindingID - proxy id of proxy bound to input connection
		input - mod 1000 value of Input Connection ID
--]]
function NOTIFY.INPUT_CHANGED(bindingID, input)
    local tParams = {INPUT = input}
    SendNotify("INPUT_CHANGED", tParams, bindingID)
end
------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
	SendSimpleNotify("ON")
end

function NOTIFY.OFF()
	SendSimpleNotify("OFF")
end
------------------------------ Volume Notification Code ------------------------------
--[[
	Notify: VOLUME_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Volume level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 level uses a percentage scale: 0 - 100
--]]
function NOTIFY.VOLUME_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("VOLUME_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BASS_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Bass level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BASS_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BASS_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: TREBLE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Treble level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.TREBLE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("TREBLE_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BALANCE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Balance level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BALANCE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BALANCE_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: LOUNDENSS_CHANGED
	Sent to the Projector proxy to indicate that the Loudness state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.LOUDNESS_CHANGED(bindingID, output, state)
    local tParams = {LOUDNESS = state, OUTPUT = output + 4000}
    SendNotify("LOUDNESS_CHANGED", tParams, bindingID)		
end

--[[
	Notify: MUTE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Mute state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.MUTE_CHANGED(bindingID, output, state)
    local tParams = {MUTE = state, OUTPUT = output + 4000}
    SendNotify("MUTE_CHANGED", tParams, bindingID)		
end

function NOTIFY.CHANNEL_CHANGED(bindingID, channel)
    local tParams = {CHANNEL = channel}
    SendNotify("CHANNEL_CHANGED", tParams, bindingID)	 
end		 end)
package.preload['projector.projector_proxy_commands'] = (function (...)
--[[=============================================================================
    ReceivedFromProxy Code for the Projector Proxy

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.ON(idBinding, tParams)
	gProjectorProxy:prx_ON(tParams)
end

function PRX_CMD.OFF(idBinding, tParams)
	gProjectorProxy:prx_OFF(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gProjectorProxy:prx_SET_INPUT(idBinding, tParams)
end

function PRX_CMD.TV_VIDEO(idBinding, tParams)
	gProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	gProjectorProxy:prx_MUTE_OFF(tParams)
end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	gProjectorProxy:prx_MUTE_ON(tParams)	
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	gProjectorProxy:prx_MUTE_TOGGLE(tParams)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	gProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_UP(tParams)
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_DOWN(tParams)
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_UP(tParams)
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_DOWN(tParams)
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_UP(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.INFO(idBinding, tParams)
	gProjectorProxy:prx_INFO(idBinding, tParams)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	gProjectorProxy:prx_GUIDE(idBinding, tParams)
end

function PRX_CMD.MENU(idBinding, tParams)
	gProjectorProxy:prx_MENU(idBinding, tParams)
end

function PRX_CMD.CANCEL(idBinding, tParams)
	gProjectorProxy:prx_CANCEL(idBinding, tParams)
end

function PRX_CMD.UP(idBinding, tParams)
	gProjectorProxy:prx_UP(idBinding, tParams)
end

function PRX_CMD.DOWN(idBinding, tParams)
	gProjectorProxy:prx_DOWN(idBinding, tParams)
end

function PRX_CMD.LEFT(idBinding, tParams)
	gProjectorProxy:prx_LEFT(idBinding, tParams)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	gProjectorProxy:prx_RIGHT(idBinding, tParams)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_DOWN(idBinding, tParams)
end

function PRX_CMD.START_UP(idBinding, tParams)
	gProjectorProxy:prx_START_UP(idBinding, tParams)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	gProjectorProxy:prx_START_LEFT(idBinding, tParams)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_START_RIGHT(idBinding, tParams)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_UP(idBinding, tParams)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	gProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
end

function PRX_CMD.ENTER(idBinding, tParams)
	gProjectorProxy:prx_ENTER(idBinding, tParams)
end

function PRX_CMD.RECALL(idBinding, tParams)
	gProjectorProxy:prx_RECALL(idBinding, tParams)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
	gProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.NUMBER_0(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_0(idBinding, tParams)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_1(idBinding, tParams)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_2(idBinding, tParams)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_3(idBinding, tParams)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_4(idBinding, tParams)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_5(idBinding, tParams)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_6(idBinding, tParams)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_7(idBinding, tParams)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_8(idBinding, tParams)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_9(idBinding, tParams)
end

function PRX_CMD.STAR(idBinding, tParams)
	gProjectorProxy:prx_STAR(idBinding, tParams)
end

function PRX_CMD.POUND(idBinding, tParams)
	gProjectorProxy:prx_POUND(idBinding, tParams)
end

function PRX_CMD.HYPHEN(idBinding, tParams)
	gProjectorProxy:prx_HYPHEN(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.PULSE_CH_UP(idBinding, tParams)
	gProjectorProxy:prx_PULSE_CH_UP(idBinding, tParams)
end

function PRX_CMD.START_CH_UP(idBinding, tParams)
	gProjectorProxy:prx_START_CH_UP(idBinding, tParams)
end

function PRX_CMD.PULSE_CH_DOWN(idBinding, tParams)
	gProjectorProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
end

function PRX_CMD.START_CH_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_CH_DOWN(idBinding, tParams)
end

function PRX_CMD.SET_CHANNEL(idBinding, tParams)
	gProjectorProxy:prx_SET_CHANNEL(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Transport Navigation Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.PLAY(idBinding, tParams)
	gProjectorProxy:prx_PLAY(idBinding, tParams)
end
	
function PRX_CMD.PAUSE(idBinding, tParams)
	gProjectorProxy:prx_PAUSE(idBinding, tParams)
end

function PRX_CMD.STOP(idBinding, tParams)
	gProjectorProxy:prx_STOP(idBinding, tParams)
end

function PRX_CMD.SKIP_REV(idBinding, tParams)
	gProjectorProxy:prx_SKIP_REV(idBinding, tParams)
end

function PRX_CMD.SKIP_FWD(idBinding, tParams)
	gProjectorProxy:prx_SKIP_FWD(idBinding, tParams)
end

function PRX_CMD.SCAN_REV(idBinding, tParams)
	gProjectorProxy:prx_SCAN_REV(idBinding, tParams)
end

function PRX_CMD.SCAN_FWD(idBinding, tParams)
	gProjectorProxy:prx_SCAN_FWD(idBinding, tParams)
end

function PRX_CMD.START_SCAN_REV(idBinding, tParams)
	gProjectorProxy:prx_START_SCAN_REV(idBinding, tParams)
end

function PRX_CMD.STOP_SCAN_REV(idBinding, tParams)
	gProjectorProxy:prx_STOP_SCAN_REV(idBinding, tParams)
end

function PRX_CMD.START_SCAN_FWD(idBinding, tParams)
	gProjectorProxy:prx_START_SCAN_FWD(idBinding, tParams)
end
function PRX_CMD.STOP_SCAN_FWD(idBinding, tParams)
	gProjectorProxy:prx_STOP_SCAN_FWD(idBinding, tParams)
end

function PRX_CMD.RECORD(idBinding, tParams)
	gProjectorProxy:prx_RECORD(idBinding, tParams)
end end)
package.preload['projector.projector_proxy_class'] = (function (...)
--[[=============================================================================
    Projector Proxy Class Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

DEFAULT_POWER_STATUS_QUERY_INTERVAL = 3         	-- Amount of time to wait BEFORE sending the next power status query, in seconds

ProjectorProxy = inheritsFrom(nil)

function ProjectorProxy:construct(projectorBindingID, bProcessesDeviceMessages, tVolumeRampingTracking, bUsePulseCommandsForVolumeRamping)
	-- member variables
	self._ProjectorBindingID = projectorBindingID
	self._PowerState = ""							--Valid Values: "ON", "OFF", "POWER_ON_SEQUENCE", "POWER_OFF_SEQUENCE"
	self._VolumeIsRamping = false
	self._VolumeRamping = tVolumeRampingTracking	--{["state"] = false, ["mode"] = "",} ||	"state" is boolean, "mode" values: "VOLUME_UP" & "VOLUME_DOWN"
	self._UsePulseCommandsForVolumeRamping = bUsePulseCommandsForVolumeRamping or false
	self._LastVolumeStatusValue = 0	
	self._MenuNavigationInProgress = false
     self._MenuNavigationMode = ""
	self._MenuNavigationProxyID = ""	
	self._CurrentlySelectedInput = ""
	self._ProcessesDeviceMessages = bProcessesDeviceMessages
	self._ControlMethod = ""						--Valid Values: "NETWORK", "SERIAL", "IR" 
	self._PowerStatusQueryDelayS = DEFAULT_COMMAND_RESPONSE_INTERVAL
end

function ProjectorProxy:Initialize(PowerStatusQueryDelayInterval, CallbackParam)
	if (PowerStatusQueryDelayInterval ~= nil) then
		self._PowerStatusQueryDelayS = PowerStatusQueryDelayInterval
	end
	self._PowerStatusQueryDelayTimer = c4_timer:new("PowerStatusQueryDelay", self._PowerStatusQueryDelayS, "SECONDS", ProjectorProxy.OnPowerStatusQueryDelayTimerExpired, false, self)
end

-- Note the '.' instead of the ':'
function ProjectorProxy.OnPowerStatusQueryDelayTimerExpired(Instance)
	LogTrace("Power Status Query Timer expired")
	Instance._PowerStatusQueryDelayTimer:KillTimer()
	
	if (GetDevicePowerStatus ~= nil and type(GetDevicePowerStatus) == "function") then
		GetDevicePowerStatus(true)  --"true" sends command as a priority command
	end
end

------------------------------------------------------------------------
-- Projector Proxy Commands (PRX_CMD)
------------------------------------------------------------------------
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_ON(tParams)
--[[
	if (self._PowerState ~= nil) then
		if (self._PowerState == 'ON') or (self._PowerState == 'POWER_ON_SEQUENCE') then 
			LogInfo("Power State is '" .. self._PowerState .. "', exiting ON")
			return 
		end	
	end	
	if (self._ProcessesDeviceMessages == false) then
		self._PowerState = 'ON'
	else	
		self._PowerState = 'POWER_ON_SEQUENCE'
	end	
--]]

    if (gControlMethod == "IR") then			
		NOTIFY.ON()			
    end	
    
    ON()
end

function ProjectorProxy:prx_OFF(tParams)
    if (self._ProcessesDeviceMessages == false) then
	   self._PowerState = 'OFF'
    else	
	   self._PowerState = 'POWER_OFF_SEQUENCE'
    end	
    if (gControlMethod == "IR") then			
	   NOTIFY.OFF()			
    end 	
    OFF()
  
    self._CurrentlySelectedInput = -1
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_SET_INPUT(idBinding, tParams)
	local input = tonumber(tParams["INPUT"] % 1000)
	self._CurrentlySelectedInput = input
	if (gControlMethod == "IR") then			
		NOTIFY.INPUT_CHANGED(self._ProjectorBindingID, input)				
	end 		

	SET_INPUT(idBinding, input)
end

function ProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "TV_VIDEO")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_MUTE_OFF(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "False")
	end 		
	MUTE_OFF()
end

function ProjectorProxy:prx_MUTE_ON(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "True")
	end 		
	MUTE_ON()
end

function ProjectorProxy:prx_MUTE_TOGGLE(tParams)
	MUTE_TOGGLE()
end

function ProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
	local c4VolumeLevel = tonumber(tParams['LEVEL'])
	SET_VOLUME_LEVEL(c4VolumeLevel)
end

function ProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
	PULSE_VOL_DOWN()
end

function ProjectorProxy:prx_PULSE_VOL_UP(tParams)
	PULSE_VOL_UP()
end

function ProjectorProxy:prx_START_VOL_DOWN(tParams)
	self:ChangeVolume("START_VOL_DOWN")
end

function ProjectorProxy:prx_START_VOL_UP(tParams)
	self:ChangeVolume("START_VOL_UP")
end

function ProjectorProxy:prx_STOP_VOL_DOWN(tParams)
	self:ChangeVolume("STOP_VOL_DOWN")
end

function ProjectorProxy:prx_STOP_VOL_UP(tParams)
	self:ChangeVolume("STOP_VOL_UP")
end

---------------------- Volume Helper Functions ----------------------
function ProjectorProxy:ChangeVolume(command_name)
	if (command_name == "STOP_VOL_UP") or (command_name == "STOP_VOL_DOWN") then
		self._VolumeIsRamping = false
		self._VolumeRamping.state = false
		self._VolumeRamping.mode = ""
	elseif (command_name == "START_VOL_UP") then 
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_UP" 
		PULSE_VOL_UP()	
	elseif (command_name == "START_VOL_DOWN") then 	
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_DOWN"	
		PULSE_VOL_DOWN()		
	else
		LogWarn(command_name .. " not handled in ChangeVolume()")
	end
end

function ProjectorProxy:ContinueVolumeRamping()
	--local command
	--if (gControlMethod == "IR") then   
		if (self._VolumeRamping.mode == "VOLUME_UP") then
			PULSE_VOL_UP()	
		elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
			PULSE_VOL_DOWN()	
		else
			LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		end	
	--else
	--[[
		local volume = self._LastVolumeStatusValue
		local deviceVolumeLevel = self:GetNextVolumeCurveValue(volume)
		if (deviceVolumeLevel ~= nil) then
			self._LastVolumeStatusValue = deviceVolumeLevel
			SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)                                 
		else
			LogWarn("ContinueVolumeRamping() next value is nil")
			return
		end
     --]]
	--end 
end

function ProjectorProxy:GetNextVolumeCurveValue(volume)
	local i, point
	volume=tonumber(volume)
	if (self._VolumeRamping.mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		LogWarn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ProjectorProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    local curveV = {}
    curveV.__index = curveV

    function curveV:new(min, max, base)
	 local len = max-min
	 local logBase = math.log(base)
	 local baseM1 = base-1

	 local instance = {
	   min = min,
	   max = max,
	   len = len,
	   base = base,
	   logBase = logBase,
	   baseM1 = baseM1,
	   toNormal = function(x)
		return (x-min)/len
	   end,
	   fromNormal = function(x)
		return min+x*len
	   end,
	   value = function(x)
		return math.log(x*baseM1+1)/logBase
	   end,
	   invert = function(x)
		return (math.exp(x*logBase)-1)/baseM1
	   end
	 }
	 return setmetatable(instance, self)
    end

    function curveV:list(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))

	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	 end
    end
    
    function curveV:create_curve(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))
	 
      local tCurve = {}
	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	   --print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	    
	    local x = math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    
	    if (has_value(tCurve, x) == false) then
		  table.insert(tCurve, x)
		  print(i, x)
	    end
	 end
	 
	 return tCurve
    end
    
    

    -- min, max, base of log (must be > 1), try some for the best results
    local a = curveV:new(minDeviceLevel, maxDeviceLevel, slope)

    -- from, to, steps
    --a:list(minDeviceLevel, maxDeviceLevel, steps)
    
    local t = a:create_curve(minDeviceLevel, maxDeviceLevel, steps)
    
    return t

end

function has_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return true
        end
    end

    return false
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	  local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --LogInfo("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --LogInfo("vl = " .. vl)
	  vl= tonumber_loc(("%.".."0".."f"):format(vl))
	  --LogInfo("vl new = " .. vl)
	  LogInfo("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_PULSE_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_START_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_START_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_PRESET_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_UP")
end

function ProjectorProxy:prx_PRESET_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_DOWN")
end

function ProjectorProxy:prx_SET_CHANNEL(idBinding, tParams)
    local channel= tParams['CHANNEL']
	SET_CHANNEL(idBinding, channel)
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_INFO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "INFO")
end

function ProjectorProxy:prx_GUIDE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "GUIDE")
end

function ProjectorProxy:prx_MENU(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "MENU")
end

function ProjectorProxy:prx_CANCEL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "CANCEL")
end

function ProjectorProxy:prx_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "UP")
end

function ProjectorProxy:prx_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "DOWN")
end

function ProjectorProxy:prx_LEFT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "LEFT")
end

function ProjectorProxy:prx_RIGHT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RIGHT")
end

function ProjectorProxy:prx_START_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_DOWN")
end

function ProjectorProxy:prx_START_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_UP")
end

function ProjectorProxy:prx_START_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_LEFT")
end

function ProjectorProxy:prx_START_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_RIGHT")
end

function ProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_DOWN")
end

function ProjectorProxy:prx_STOP_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_UP")
end

function ProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_LEFT")
end

function ProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_RIGHT")
end

function ProjectorProxy:prx_ENTER(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "ENTER")
end

function ProjectorProxy:prx_RECALL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RECALL")
end

function ProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "OPEN_CLOSE")
end

function ProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_A")
end

function ProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_B")
end

function ProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_C")
end

function ProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_D")
end

---------------------- Menu Navigation Helper Functions ----------------------
function ProjectorProxy:NavigateMenu(idBinding, command_name)
	if (command_name == "STOP_UP") 
				or (command_name == "STOP_DOWN") 
				or (command_name == "STOP_LEFT") 
				or (command_name == "STOP_RIGHT") then
		self._MenuNavigationInProgress = false
		self._MenuNavigationMode = ""
		self._MenuNavigationProxyID = ""
		return
	elseif (command_name == "START_UP") then 
		self._MenuNavigationMode = "UP"
	elseif (command_name == "START_DOWN") then 	
		self._MenuNavigationMode = "DOWN"	
	elseif (command_name == "START_LEFT") then 
		self._MenuNavigationMode = "LEFT"	
     elseif (command_name == "START_RIGHT") then 
		self._MenuNavigationMode = "RIGHT"
	else
		LogWarn(command_name .. " not handled in NavigateMenu()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, self._MenuNavigationMode)
	self._MenuNavigationInProgress = true
	self._MenuNavigationProxyID = idBinding
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_NUMBER_0(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_0")
end

function ProjectorProxy:prx_NUMBER_1(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_1")
end

function ProjectorProxy:prx_NUMBER_2(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_2")
end

function ProjectorProxy:prx_NUMBER_3(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_3")
end

function ProjectorProxy:prx_NUMBER_4(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_4")
end

function ProjectorProxy:prx_NUMBER_5(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_5")
end

function ProjectorProxy:prx_NUMBER_6(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_6")
end

function ProjectorProxy:prx_NUMBER_7(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_7")
end

function ProjectorProxy:prx_NUMBER_8(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_8")
end

function ProjectorProxy:prx_NUMBER_9(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_9")
end

function ProjectorProxy:prx_STAR(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "STAR")
end

function ProjectorProxy:prx_POUND(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "POUND")
end

function ProjectorProxy:prx_HYPHEN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "HYPHEN")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Transport Navigation Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_PLAY(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PLAY")
end
	
function ProjectorProxy:prx_PAUSE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PAUSE")
end

function ProjectorProxy:prx_STOP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "STOP")
end

function ProjectorProxy:prx_SKIP_REV(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "SKIP_REV")
end

function ProjectorProxy:prx_SKIP_FWD(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "SKIP_FWD")
end

function ProjectorProxy:prx_SCAN_REV(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "SCAN_REV")
end

function ProjectorProxy:prx_SCAN_FWD(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "SCAN_FWD")
end

function ProjectorProxy:prx_START_SCAN_REV(idBinding, tParams)
	self:NavigateTransport(idBinding, "START_SCAN_REV")
end

function ProjectorProxy:prx_STOP_SCAN_REV(idBinding, tParams)
	self:NavigateTransport(idBinding, "STOP_SCAN_REV")
end

function ProjectorProxy:prx_START_SCAN_FWD(idBinding, tParams)
	self:NavigateTransport(idBinding, "START_SCAN_FWD")
end

function ProjectorProxy:prx_STOP_SCAN_FWD(idBinding, tParams)
	self:NavigateTransport(idBinding, "STOP_SCAN_FWD")
end

function ProjectorProxy:prx_RECORD(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RECORD")
end

function ProjectorProxy:NavigateTransport(idBinding, command_name)
	if (command_name == "STOP_SCAN_REV") 
				or (command_name == "STOP_SCAN_FWD") then
		self._TransportNavigationInProgress = false
		self._TransportNavigationMode = ""
		self._TransportNavigationProxyID = ""
		return
	elseif (command_name == "START_SCAN_REV") then 
		self._TransportNavigationMode = "SCAN_REV"
	elseif (command_name == "START_SCAN_FWD") then 	
		self._TransportNavigationMode = "SCAN_FWD"	
	else
		LogWarn(command_name .. " not handled in NavigateTransport()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, self._TransportNavigationMode)
	self._TransportNavigationInProgress = true
	self._TransportNavigationProxyID = idBinding
	
end

------------------------------------------------------------------------
-- Projector Proxy Notifies
------------------------------------------------------------------------

function ProjectorProxy:dev_InputChanged(bindingID, input)
	NOTIFY.INPUT_CHANGED(bindingID, input)
end

function ProjectorProxy:dev_PowerOn()
	self._PowerState = "ON"
	NOTIFY.ON()	
end

function ProjectorProxy:dev_PowerOff()
	self._PowerState = "OFF"
	NOTIFY.OFF()
end

function ProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	NOTIFY.VOLUME_LEVEL_CHANGED(self._ProjectorBindingID, output, c4Level)	
	
	if (self._VolumeIsRamping) then
		--do nothing
		--during volume ramping, LastVolumeStatusValue is set in ContinueVolumeRamping()
	else
		self._LastVolumeStatusValue = deviceLevel
	end	
end

function ProjectorProxy:dev_MuteChanged(output, state)
	NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, output, state)
end		

function ProjectorProxy:dev_BassLevelChanged(output, level)
	NOTIFY.BASS_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_TrebleLevelChanged(output, level)
	NOTIFY.TREBLE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_BalanceLevelChanged(output, level)
	NOTIFY.BALANCE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_LoudnessChanged(output, state)
	NOTIFY.LOUDNESS_CHANGED(self._ProjectorBindingID, output, state)
end

function ProjectorProxy:dev_ChannelStatus(channel)
	NOTIFY.CHANNEL_CHANGED(self._ProjectorBindingID, channel)
end end)
package.preload['common.c4_url_connection'] = (function (...)
--[[=============================================================================
    Base for a url connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_url_connection = "2016.01.08"
end

UrlConnectionBase = inheritsFrom(DeviceConnectionBase)

function UrlConnectionBase:construct(Url)
	self.superClass():construct()
	self._Url = Url
end

function UrlConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "URL"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	OnURLConnectionChanged()
end

function UrlConnectionBase:ControlMethod()
	return "URL"
end

function UrlConnectionBase:SetUrl(Url)
	self._Url = Url
end

function UrlConnectionBase:SendCommand(sCommand, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(self._Url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(self._Url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:SendCommandUrl(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlPost(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlGet(url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlGet(url, sHeader)
		else
			ticketId = C4:urlGet(url)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:ReceivedAsync(ticketId, sData, responseCode, tHeaders)
	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(sData))
	LogTrace(tHeaders)
	local tMessage = {
		["ticketId"] = ticketId,
		["sData"] = sData,
		["responseCode"] = responseCode,
		["tHeaders"] = tHeaders
	}
	
	status, err = pcall(HandleMessage, tMessage)
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

function ConnectURL()
	gIsUrlConnected = true
	SetControlMethod()
end

function DisconnectURL()
	gIsUrlConnected = false
	SetControlMethod()
end
 end)
package.preload['connections'] = (function (...)
--[[=============================================================================
    Functions for managing the status of the drivers bindings and connection state

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_url_connection"

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.connections = "2015.03.31"
end

-- constants
COM_USE_ACK = true
COM_COMMAND_DELAY_MILLISECONDS = 100
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4

NETWORK_PORT = 9761
IR_BINDING_ID = 2
SERIAL_BINDING_ID = 1
NETWORK_BINDING_ID = 6001
BROADCAST_BINDING_ID = 6002
BROADCAST_PORT = 9

--[[=============================================================================
    OnSerialConnectionChanged(idBinding, class, bIsBound)

    Description:
    Function called when a serial binding changes state(bound or unbound).

    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.

    Returns:
    None
===============================================================================]]
function OnSerialConnectionChanged(idBinding, class, bIsBound)

end

--[[=============================================================================
    OnIRConnectionChanged(idBinding, class, bIsBound)

    Description:
    Function called when an IR binding changes state(bound or unbound).

    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.

    Returns:
    None
===============================================================================]]
function OnIRConnectionChanged(idBinding, class, bIsBound)

end

--[[=============================================================================
    OnNetworkConnectionChanged(idBinding, bIsBound)

    Description:
    Function called when a network binding changes state(bound or unbound).

    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.

    Returns:
    None
===============================================================================]]
function OnNetworkConnectionChanged(idBinding, bIsBound)
    if (idBinding == NETWORK_BINDING_ID) and bIsBound then
	   gNetworkInitTimer:StartTimer()
    elseif (idBinding == NETWORK_BINDING_ID) and (bIsBound ~= true) then

    end	
end

function OnNetworkInitTimerExpired()
    gNetworkInitTimer:KillTimer()
    GetMACAddressWired()
    GetMACAddressWifi()
end

--[[=============================================================================
    OnNetworkStatusChanged(idBinding, nPort, sStatus)

    Description:
    Called when the network connection status changes. Sets the updated status of the specified binding

    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline

    Returns:
    None
===============================================================================]]
function OnNetworkStatusChanged(idBinding, nPort, sStatus)
    if (idBinding == NETWORK_BINDING_ID) then
	   if (sStatus	== "ONLINE") then
		  if (Properties["Wired MAC Address"] == nil or Properties["Wired MAC Address"] == "" or Properties["Wi-Fi MAC Address"] == nil or Properties["Wi-Fi MAC Address"] == "" ) then
			 gNetworkInitTimer:StartTimer()
		  end
		  
		  --if (gProjectorProxy._PowerState == "OFF") then gQueryPowerDelayTimer:StartTimer(2, "SECONDS") end
		  
	   else
		  --TV does not respond when OFF
		  --if timer is not killed here, then the driver will continually try to reconnect when TV is OFF
		  gCon._NetworkReconnectTimer:KillTimer()	
		  
		  --TV may have been turned off by factory remote or due to lack of signal...
		  if (gProjectorProxy._PowerState ~= "OFF") then
			 gProjectorProxy:dev_PowerOff()
		  end
	   end				 
    else
    end	
end

--[[=============================================================================
    OnURLConnectionChanged(url)

    Description:
    Function called when the c4_url_connection is created.

    Parameters:
    url - url used by the url connection.

    Returns:
    None
===============================================================================]]
function OnURLConnectionChanged(url)

end

--[[=============================================================================
    DoEvents()

    Description:
    Called from OnSendTimeExpired in the DeviceConnectionBase.

    Parameters:
    None

    Returns:
    None
===============================================================================]]

gWasVolumeRamping = false
gWasMenuNavigating = false
gWasTransportNavigating = false
function DoEvents()
    if (gProjectorProxy._VolumeIsRamping == true) then
	   if (gProjectorProxy._VolumeRamping.state == true) then 
		  if (gControlMethod == "Network") then 
			 gProjectorProxy:ContinueVolumeRamping()
			 gWasVolumeRamping = true
		  else
			 gVolumeRampTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS") 
		  end
	   end
    else
	   if (gWasVolumeRamping) then
		  --note from Peter: not sure I like this approach but this is a test to clear any remaining volume commands in the queue
		  --there is an outside chance that other non-volume commands may have been queued and these will be lost. not sure of the likelihood though...
		  gCon._CommandQueue:clear()
		  gWasVolumeRamping = false
	   end
    end
	
    if (gProjectorProxy._MenuNavigationInProgress == true) then
	   --SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._MenuNavigationProxyID, gProjectorProxy._MenuNavigationMode)
	   --gWasMenuNavigating = true
	   if (gControlMethod == "Network") then 
		  SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._MenuNavigationProxyID, gProjectorProxy._MenuNavigationMode)
		  gWasMenuNavigating = true
	   else
		  gMenuNavigationTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS")
	   end
    else
	   if (gWasMenuNavigating) then
		  gCon._CommandQueue:clear()
		  gWasMenuNavigating = false
	   end
    end		
    
    if (gProjectorProxy._TransportNavigationInProgress == true) then
	   --SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._MenuNavigationProxyID, gProjectorProxy._MenuNavigationMode)
	   --gWasMenuNavigating = true
	   if (gControlMethod == "Network") then 
		  SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._TransportNavigationProxyID, gProjectorProxy._TransportNavigationMode)
		  gWasTransportNavigating = true
	   else
		  gTransportNavigationTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS")
	   end
    else
	   if (gWasTransportNavigating) then
		  gCon._CommandQueue:clear()
		  gWasTransportNavigating = false
	   end
    end	
end

function OnVolumeRampTimerExpired(command)
    LogTrace("OnVolumeRampTimerExpired()")
    gVolumeRampTimer:KillTimer()
    if (gProjectorProxy._VolumeRamping.state == true) then
	   gProjectorProxy:ContinueVolumeRamping()
	   gWasVolumeRamping = true
    end
end

function OnMenuNavigationTimerExpired(command)
    LogTrace("OnMenuNavigationTimerExpired()")
    gMenuNavigationTimer:KillTimer()
    SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._MenuNavigationProxyID, gProjectorProxy._MenuNavigationMode)
    gWasMenuNavigating = true
end

function OnTransportNavigationTimerExpired(command)
    LogTrace("OnTransportNavigationTimerExpired()")
    gTransportNavigationTimer:KillTimer()
    SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._TransportNavigationProxyID, gProjectorProxy._TransportNavigationMode)
    gWasTransportNavigating = true
end


--[[=============================================================================
    SendKeepAlivePollingCommand()

    Description:
    Sends a driver specific polling command to the connected system

    Parameters:
    None

    Returns:
    None
===============================================================================]]
function SendKeepAlivePollingCommand()
    LogTrace("SendKeepAlivePollingCommand()")

	--TODO: set keep alive command for the network connected system if required.
    local command = CMDS_IP["GET_MUTE_STATUS"]
    LogTrace("command = " .. command)
    PackAndQueueCommand("SendKeepAlivePollingCommand", command)
end
 end)
package.preload['proxy_commands'] = (function (...)
--[[=============================================================================
    Commands received from the TV proxy (ReceivedFromProxy)

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: ON
	Parameters:
--]]
function ON()
    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
          
    if (gProjectorProxy._PowerState ~= nil) then
	   if (gProjectorProxy._PowerState == 'ON')  then 
		  if ((gControlMethod == "IR") or (gProjectorProxy._ProcessesDeviceMessages == false)) then
			 command_delay = tonumber(Properties["Power On Delay Seconds"])
			 delay_units = "SECONDS"
		  else
			 command_delay = tonumber(Properties["Command Delay Milliseconds"])
			 delay_units = "MILLISECONDS"
		  end
	   elseif (gProjectorProxy._PowerState == 'POWER_ON_SEQUENCE') then
		  command_delay = tonumber(Properties["Power On Delay Seconds"])
		  delay_units = "SECONDS"
	   end	
    else
	   command_delay = tonumber(Properties["Power On Delay Seconds"])
	   delay_units = "SECONDS"
    end	

    gPowerOnTimer:StartTimer(30, delay_units)
    
    if gPowerOffTimer:TimerStarted() then
	   --quick power on AFTER power off, need TV to turn off before sending on. This occurs in OnPowerOffTimer...
	   return
    end
    
    if (gControlMethod == nil) then sendWOL() end	--AFTER DRIVER UPDATED gControlMethod may not be set...
    
    local command
    if (gControlMethod == "IR") then
	   command = CMDS_IR["ON"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("ON", command, command_delay, delay_units)
		
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["ON"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("ON", command, command_delay, delay_units)
	   PackAndQueueCommand("ON", command, command_delay, delay_units) -- LG send twice to be sure
		
    elseif (gControlMethod == "Network") then
	   gQueryPowerDelayTimer:StartTimer(command_delay, delay_units)
	   sendWOL()
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return

    end
    
end

function sendWOL()
    if (LuaC4Object.SendWOL) then	--C4:SendWOL() added in OS 2.10
	   local mac_string = format_mac_string()
	   
	   for i = 1,5 do	   
		  C4:SendWOL(mac_string, 9)
	   end
    else
	   local mac_hex = format_mac_hex()
	   C4:CreateNetworkConnection (BROADCAST_BINDING_ID, "255.255.255.255", BROADCAST_PORT) --create network connection BROADCAST_BINDING_ID for Wake on Lan
	   C4:NetConnect(BROADCAST_BINDING_ID, BROADCAST_PORT, "UDP") --connect to connection BROADCAST_BINDING_ID and make it UDP for Wake On Lan
	   
	   local packet = string.rep(string.char(255), 6) .. string.rep(mac_hex, 16) -- Build 'magic packet'. 
	   --hexdump (packet) 
	   for i = 1,5 do	  
		  C4:SendToNetwork (BROADCAST_BINDING_ID, BROADCAST_PORT, packet)
	   end
    end
    
    gWoLNetConnectDelayTimer:StartTimer(2, "SECONDS")
end

function format_mac_string()
     local s = ""
     if (Properties["Network Connection"] == "Wi-Fi") then
	   s = Properties["Wi-Fi MAC Address"]
	else
	   s = Properties["Wired MAC Address"]
	end
	
	local mac = s:gsub("%X", "")
	return mac
end

function format_mac_hex()
     local s = ""
     if (Properties["Network Connection"] == "Wi-Fi") then
	   s = Properties["Wi-Fi MAC Address"]
	else
	   s = Properties["Wired MAC Address"]
	end
	
	local mac = s:gsub("%X", " ")
	return tohex(mac)
end

--[[
	Proxy Command: OFF
	Parameters:
--]]
function OFF()
    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    
    gPowerOnTimer:KillTimer()
    gPowerOffTimer:StartTimer(command_delay, delay_units)
    
    local command
    if (gControlMethod == "IR") then
		command = CMDS_IR["OFF"]
		LogTrace("command = " .. command)
		PackAndQueueCommand("OFF", command, command_delay, delay_units)
    elseif (gControlMethod == "Serial") then
		command = CMDS_Serial["OFF"]
		LogTrace("command = " .. command)
		PackAndQueueCommand("OFF", command, command_delay, delay_units)
		PackAndQueueCommand("OFF", command, command_delay, delay_units)	-- LG end twice to be sure
    elseif (gControlMethod == "Network") then
	   command = CMDS_IP["OFF"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("OFF", command, command_delay, delay_units)
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end
    
    gMuted = ""
    -- TODO: If the device will automatically report power status after
    --	the Off command is sent, then the line below can be commented out
    --GetDevicePowerStatus()
    
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: SET_INPUT
	Parameters:
		idBinding: proxybindingid of proxy bound to input connection
		output: mod 1000 value of Output Connection id
		input: mod 1000 value of Input Connection id
--]]
function SET_INPUT(idBinding, input)
    
    --for Network & Serial control. IR will use default command delay...
    local delay, delay_units
    if (gPowerOnTimer:TimerStarted()) then
	   delay = 30	--tonumber(Properties["Power On Delay Seconds"])
	   delay_units = "SECONDS"
    else
	   delay = tonumber(Properties["Command Delay Milliseconds"])
	   delay_units = "MILLISECONDS"
    end
    
    local command
    if (gControlMethod == "IR") then
	   command = tInputCommandMap_IR[tInputConnMapByID[input].Name]
	   PackAndQueueCommand("SET_INPUT", command)
    elseif (gControlMethod == "Serial") then
	   command = tInputCommandMap_Serial[tInputConnMapByID[input].Name]
	   --PackAndQueueCommand("SET_INPUT", command)
	   gSetInputDelayTimer:StartTimer(delay, delay_units)
	   gSetInputDelayTimer._CallbackParam = command	  
    elseif (gControlMethod == "Network") then
    
	   if (input >= MINIAPP_BINDING_START % 1000 and input <= MINIAPP_BINDING_END % 1000) then
		  -- Get the device ID of the proxy handling the miniapp switch on this driver
		  local proxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), idBinding))

		  -- Get the device ID of the minidriver proxy connected to the requested input on this driver
		  local appProxyId = C4:GetBoundProviderDevice (proxyDeviceId, input + 3000)

		  -- Get the device ID of the minidriver protocol connected to the minidriver proxy
		  local appDeviceId = C4:GetBoundProviderDevice (appProxyId, 5001)

		  -- get the details for the app for this kind of universal-minidriver-compatible type
		  local appId = GetRelevantUniversalAppId (appDeviceId, MINIAPP_TYPE)
		  local appName = GetRelevantUniversalAppId (appDeviceId, 'APP_NAME')

		  command = appId
		  
		  --update IO mapping tables
		  local id = string.gsub(appId, 'APP_LAUNCH ', '')
		  tInputResponseMap_IP[id] = appName
		  tInputConnMapByID[input] = {Name = appName,BindingID = 5001,}
		  tInputConnMapByName[appName] = {ID = input,BindingID = 5001,}

		  if ((Properties ['Passthrough Mode'] or 'On') ~= 'On') then
			  local passthroughProxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), PASSTHROUGH_PROXY))
			  local _timer = function (timer)
				  print ('Looking for ' .. appProxyId)
				  for roomId, deviceId in pairs (RoomIDSources) do
					  if (deviceId == appProxyId) then
						  C4:SendToDevice (roomId, 'SELECT_VIDEO_DEVICE', {deviceid = passthroughProxyDeviceId})
					  end
				  end
			  end
			  C4:SetTimer (500, _timer)
		  end   
		  
	   else
		  command = tInputCommandMap_IP[tInputConnMapByID[input].Name]
	   end
    
	   gSetInputDelayTimer:StartTimer(delay, delay_units)
	   gSetInputDelayTimer._CallbackParam = command	   
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end

    --needed for SET_CHANNEL function...
    set_channel_type(input)
end

function set_channel_type(input)
    --check to see if input being selected is an RF input. If so, then set global
    if (input >= 101 and input <= 111) then
	   gCurrentlySelectedRFInputName = tInputConnMapByID[input].Name
    end
end

function OnSetInputDelayTimerExpired(command)
    LogTrace("OnSetInputDelayTimerExpired, command = " .. command)
    gSetInputDelayTimer:KillTimer()
    
    PackAndQueueCommand("SET_INPUT", command, 500, "MILLISECONDS")
end

function OnQueryInputDelayTimerExpired()
    LogTrace("OnQueryInputDelayTimerExpired")
    gQueryInputDelayTimer:KillTimer()

    GetCurrentApp()
end

function OnPowerOnTimerExpired()
    LogTrace("OnPowerOnTimerExpired")
    gPowerOnTimer:KillTimer()
end

function OnPowerOffTimerExpired()
    LogTrace("OnPowerOffTimerExpired")
    gPowerOffTimer:KillTimer()
    
    gProjectorProxy:dev_PowerOff()
    
    --for quick power on after power off...
    if gPowerOnTimer:TimerStarted() then
	   gSendPowerOnAfterPowerOffDelayTimer:StartTimer(tonumber(Properties["Power Off Delay Seconds"]), "SECONDS")
    end
end

function OnSendPowerOnAfterPowerOffDelayTimerExpired()
    LogTrace("OnSendPowerOnAfterPowerOffDelayTimerExpired")
    gSendPowerOnAfterPowerOffDelayTimer:KillTimer()
    
    ON()
end

function OnWoLNetConnectDelayTimerExpired()
    LogTrace("OnWoLNetConnectDelayTimerExpired")
    gWoLNetConnectDelayTimer:KillTimer()
   
--[[   
    --hack to handle WOL not bringing network connection online...
    --caused by SDDP Notify Alive not being sent which has been remedied by LG
    if (gCon._IsOnline == false) then
	   print("WOL failed to bring network connection online. Starting NetworkReconnectTimer !!!")
	   --C4:NetDisconnect(NETWORK_BINDING_ID, NETWORK_PORT)
	   --C4:NetConnect(NETWORK_BINDING_ID, NETWORK_PORT)
	   gCon._NetworkReconnectTimer:StartTimer(1, "SECONDS")
    end
--]]
    
end

function OnQueryPowerDelayTimerExpired()
    LogTrace("OnQueryPowerDelayTimerExpired")
    gQueryPowerDelayTimer:KillTimer()
       
    GetDevicePowerStatus()
    
    if gPowerOnTimer:TimerStarted() then
	   gQueryPowerDelayTimer:StartTimer(2, "SECONDS")
    end
    
end

function OnSetChannelDelayTimerExpired(command)
    LogTrace("OnSetChannelDelayTimerExpired")
    gSetChannelDelayTimer:KillTimer()
    PackAndQueueCommand("SET_CHANNEL", command)
end

function OnQueryChannelDelayTimerExpired()
    LogTrace("OnQueryChannelDelayTimerExpired")
    gQueryChannelDelayTimer:KillTimer()
    GetCurrentChannel()
end

function OnQueryMuteDelayTimerExpired()
    LogTrace("OnQueryMuteDelayTimerExpired")
    gQueryMuteDelayTimer:KillTimer()

    GetMuteState()
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: MUTE_OFF
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function MUTE_OFF()
    local command
    if (gControlMethod == "IR") then
	   command = CMDS_IR["MUTE_OFF"]
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_OFF"]
    elseif (gControlMethod == "Network") then
	   command = CMDS_IP["MUTE_OFF"]
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end
    LogTrace("command = " .. command)
    PackAndQueueCommand("MUTE_OFF", command)
end

--[[
	Proxy Command: MUTE_ON
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function MUTE_ON()
    local command
    if (gControlMethod == "IR") then
	   command = CMDS_IR["MUTE_ON"]
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_ON"]
    elseif (gControlMethod == "Network") then
	   command = CMDS_IP["MUTE_ON"]
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end
    LogTrace("command = " .. command)
    PackAndQueueCommand("MUTE_ON", command)
end

--[[
	Proxy Command: MUTE_TOGGLE
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function MUTE_TOGGLE()
    local command
    if (gControlMethod == "IR") then
	   command = CMDS_IR["MUTE_TOGGLE"]
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_TOGGLE"]
    elseif (gControlMethod == "Network") then
	   command = CMDS_IP["MUTE_TOGGLE"]
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end
    LogTrace("command = " .. command)
    PackAndQueueCommand("MUTE_TOGGLE", command)
end

--[[
	Proxy Command: SET_VOLUME_LEVEL
	Parameters:
		output: mod 1000 value of Output Connection id
		c4VolumeLevel: volume level to be set represented in Control4 scale (0-100)
--]]
function SET_VOLUME_LEVEL(c4VolumeLevel)
     if (gControlMethod == "IR") then return end
	
	--local minDeviceLevel = 0
	--local maxDeviceLevel = 100
	local deviceVolumeLevel = c4VolumeLevel --ConvertVolumeToDevice(c4VolumeLevel, minDeviceLevel, maxDeviceLevel)

	
	--LogInfo('deviceVolumeLevel: ' .. deviceVolumeLevel)

	local command = ""
     if (gControlMethod == "Serial") then
	   deviceVolumeLevel = string.format ("%02X", deviceVolumeLevel) --convert to two byte hex string
	   command = CMDS_Serial["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
     elseif (gControlMethod == "Network") then
	   command = CMDS_IP["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
     else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
	end
	
	LogTrace("command = " .. command)
	PackAndQueueCommand("SET_VOLUME_LEVEL", command)
end

--[[
	Helper Function: SET_VOLUME_LEVEL_DEVICE
	Parameters:
		output: mod 1000 value of Output Connection id
		deviceVolumeLevel: volume level to be set represented in device scale (as sepcified in the device's control protocol)
--]]
function SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)
	--Called from ContinueVolumeRamping()

	-- TODO: create packet/command to send to the device
	deviceVolumeLevel = string.format ("%02X", deviceVolumeLevel) --convert to two byte hex string
	local command = CMDS_Serial["SET_VOLUME_LEVEL"] .. deviceVolumeLevel

	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("SET_VOLUME_LEVEL_DEVICE", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_DOWN
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function PULSE_VOL_DOWN()
	local command
	if (gControlMethod == "IR") then
		command = CMDS_IR["PULSE_VOL_DOWN"]
	elseif (gControlMethod == "Serial") then
		command = CMDS_Serial["PULSE_VOL_DOWN"]
     elseif (gControlMethod == "Network") then
	     command = CMDS_IP["PULSE_VOL_DOWN"]
     else
	     print(tostring(gControlMethod) .. " Control Method not handled!!!")
	     return
	end
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_DOWN", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_UP
	Parameters:
		output: mod 1000 value of Output Connection id
--]]
function PULSE_VOL_UP()
	local command
	if (gControlMethod == "IR") then
		command = CMDS_IR["PULSE_VOL_UP"]
	elseif (gControlMethod == "Serial") then
		command = CMDS_Serial["PULSE_VOL_UP"]
     elseif (gControlMethod == "Network") then
	     command = CMDS_IP["PULSE_VOL_UP"]
     else
	     print(tostring(gControlMethod) .. " Control Method not handled!!!")
	     return
	end
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_UP", command, command_delay)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

    gChannelTypeByInputNameMap = {}
    
    --Europe, Mid-East, Colombia, Asia except South Korea and Japan Model
    --Transmission [m][a][ ][Set ID][ ][Data 00][ ][Data 01][ ][Data 02][Cr]
    gChannelTypeByInputNameMap["Europe, Mid-East, Colombia, Asia except South Korea and Japan Model"] = {
	   ['Antenna TV (ATV)']			= '00',
	   ['Cable TV (CATV)']			= '80',
	   ['Antenna TV (DTV)']			= '10',
	   ['Cable TV (CADTV)']			= '90',
	   ['Satellite TV (SDTV)']		= '40',
	   ['Antenna Radio (Radio)']		= '20',
	   ['Satellite Radio (S-Radio)']	= '50',
	   ['Cable Radio (CA-Radio)']		= 'A0',
    }
    
    --South Korea, North/Latin America except Colombia Model
    --Transmission [m][a][ ][0][ ][Data00][ ][Data01][ ][Data02][ ][Data03][ ][Data04][ ][Data05][Cr]
    gChannelTypeByInputNameMap["South Korea, North/Latin America except Colombia Model"] = {
	   ['Antenna TV (ATV)']								= '00',
	   ['Cable TV (CATV)']								= '01',
	   ['Antenna TV (DTV)']								= '02',	--Use Physical Channel Number
	   ['Cable TV (CADTV)']								= '06',	--Use Physical Channel Number
	   ['Antenna TV (DTV)(Do Not Use Physical Channel)']		= '22',	--Don’t Use Physical Channel Number
	   ['Cable TV (CADTV)(Do Not Use Physical Channel)']		= '26',	--Don’t Use Physical Channel Number
	   ['Cable TV (CADTV)(Use Major Channel Number Only)']		= '46',	--Use Major Channel Number Only (One Part Channel)
    }
    
    --Japan Model
    --Transmission [m][a][ ][0][ ][Data00][ ][Data01][ ][Data02][ ][Data03][ ][Data04][ ][Data05][Cr]
    gChannelTypeByInputNameMap["Japan Model"] = {
	   ['Antenna TV (DTV)']				= '02',
	   ['BS (Broadcasting Satellite)']		= '07',
	   ['CS1 (Communication Satellite 1)']	= '08',
	   ['CS2 (Communication Satellite 2)']	= '09',
    }

-- Need to know current input in order to complete packet
function SET_CHANNEL(idBinding, channel)

    --for Network & Serial control. IR will use default command delay...
    local delay, delay_units
    if (gPowerOnTimer:TimerStarted()) then
	   delay = 30	--tonumber(Properties["Power On Delay Seconds"])
	   if (gSetInputDelayTimer:TimerStarted()) then delay = delay + 1 end
	   delay_units = "SECONDS"
    else
	   delay = tonumber(Properties["Command Delay Milliseconds"])
	   if (gSetInputDelayTimer:TimerStarted()) then delay = delay + 1000 end
	   delay_units = "MILLISECONDS"
    end
	   
    if (gControlMethod == "IR") then
	   for i=1,string.len(channel) do
		  local c = string.sub(channel,i,i)
		  if (string.find(c, "%D")) then
			 -- TODO: change delimiter command below if it is not a DOT...
			 SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "POUND")
		  else
			 SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_" .. c)
		  end
	   end
	   SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "ENTER")
    elseif (gControlMethod == "Network") then
    --[[
	   ['Antenna TV (ATV)']='INPUT_SELECT atv',				--EMEA, AMERICAS
	   ['Cable TV (CATV)']='INPUT_SELECT catv',				--EMEA, AMERICAS
	   ['Antenna TV (DTV)']='INPUT_SELECT dtv',				--EMEA, AMERICAS, JAPAN
	   ['Cable TV (CADTV)']='INPUT_SELECT cadtv',				--EMEA, AMERICAS    
    --]]
	   local cmd_pfx, cmd_sfx = ""
	   local major, minor = channel:match("(%d+)%D(%d*)")
	   if (major == nil) then major = channel end	--major only
	   if (gCurrentlySelectedRFInputName == "Antenna TV (ATV)") then
		  cmd_pfx = "CHANNEL_SETTING_ATSC_ATV"
		  cmd_sfx = "antenna"				--channel: major only
	   elseif (gCurrentlySelectedRFInputName == "Cable TV (CATV)") then
		  cmd_pfx = "CHANNEL_SETTING_ATSC_ATV"
		  cmd_sfx = "cable"					--channel: major only
	   elseif (gCurrentlySelectedRFInputName == "Antenna TV (DTV)") then
		  cmd_pfx = "CHANNEL_SETTING_ATSC_DTV"
		  if (minor == nil) then
			 cmd_sfx = "cablemaj"				--channel: major only
		  else
			 cmd_sfx = "antennanotphy"			--channel: major & minor
		  end
	   elseif (gCurrentlySelectedRFInputName == "Cable TV (CADTV)") then
		  cmd_pfx = "CHANNEL_SETTING_ATSC_DTV"
		  if (minor == nil) then
			 cmd_sfx = "cablemaj"				--channel: major only
		  else
			 cmd_sfx = "cablenotphy"				--channel: major & minor
		  end
		  
	   else
		  print(gCurrentlySelectedRFInputName .. " not supported in IP protocol!!!")
		  return
	   end
	   
	   local ch = ""
	   if (minor == nil) then
		  ch = " " .. major .. " "
	   else
		  ch = " " .. major .. " " .. minor .. " "
	   end
	   local command = cmd_pfx .. ch .. cmd_sfx 
	   

	   
	   gSetChannelDelayTimer:StartTimer(delay, delay_units)
	   gSetChannelDelayTimer._CallbackParam = command
	   
	   LogTrace("SET_CHANNEL, command: " .. command)

    elseif (gControlMethod == "Serial") then
	   
	   local tuner_region = Properties["Tuner Region"]
	   local channel_type = gChannelTypeByInputNameMap[tuner_region][gCurrentlySelectedRFInputName] 
	   if (channel_type == nil) then LogTrace("SET_CHANNEL, channel_type is nil!") return end
	   
	   local command, Data00, Data01, Data02, Data03, Data04, Data05, major_high_byte, major_low_byte, minor_high_byte, minor_low_byte
	   
	   local high_byte = string.format("%02X", math.floor(channel / 256))
	   local low_byte = string.format("%02X", channel % 256)
	   
	   local _, _, major =string.find(channel,'(.-)%D')
	   local _, _, minor =string.find(channel,'%D(.*)')
	   
	   if (major ~= nil) then
		  major_high_byte = string.format("%02X", math.floor(major / 256))
		  major_low_byte = string.format("%02X", major % 256)
	   else
		  major_high_byte = high_byte	
		  major_low_byte = low_byte	
	   end
	   
	   if (minor ~= nil) then
		  minor_high_byte = string.format("%02X", math.floor(minor / 256))
		  minor_low_byte = string.format("%02X", minor % 256)
	   else
		  minor_high_byte = "00"
		  minor_low_byte = "00"
	   end
	   
	   Data01 = major_high_byte
	   Data02 = major_low_byte
	   Data03 = minor_high_byte
	   Data04 = minor_low_byte
	   Data05 = channel_type
	   
	   if (tuner_region == "Europe, Mid-East, Colombia, Asia except South Korea and Japan Model") then
		  --Transmission [m][a][ ][Set ID][ ][Data 00][ ][Data 01][ ][Data 02][Cr]
		  Data00 = high_byte
		  Data01 = low_byte
		  Data02 = channel_type
		  command = CMDS_Serial["SET_CHANNEL"] .. Data00 .. ' ' .. Data01 .. ' ' .. Data02		  
		  
	   elseif (tuner_region == "South Korea, North/Latin America except Colombia Model") then
		  --Transmission [m][a][ ][0][ ][Data00][ ][Data01][ ][Data02][ ][Data03][ ][Data04][ ][Data05][Cr]
		  if (major ~= nil) then
			 --need to update channel_type since we have a valid major/minor channel
			 local modified_input_name = gCurrentlySelectedRFInputName .. "(Do Not Use Physical Channel)"
			 channel_type = gChannelTypeByInputNameMap[tuner_region][modified_input_name] 
			 if (channel_type == nil) then LogTrace("SET_CHANNEL, channel_type is nil!") return end
		  end
		  Data00 = low_byte	--for ATV & CATV, Don't care for others
		  command = CMDS_Serial["SET_CHANNEL"] .. Data00 .. ' ' .. Data01 .. ' ' .. Data02 .. ' ' .. Data03 .. ' ' .. Data04 .. ' ' .. Data05
		  
	   elseif (tuner_region == "Japan Model") then
		  --Transmission [m][a][ ][0][ ][Data00][ ][Data01][ ][Data02][ ][Data03][ ][Data04][ ][Data05][Cr]
		  Data00 = "00"
		  command = CMDS_Serial["SET_CHANNEL"] .. Data00 .. ' ' .. Data01 .. ' ' .. Data02 .. ' ' .. Data03 .. ' ' .. Data04 .. ' ' .. Data05
		  
	   else
		  LogTrace("SET_CHANNEL, Invalid Tuner Region = " .. tostring(tuner_region))
		  return
		  
	   end

	   LogTrace("command = " .. command)
	   --PackAndQueueCommand("SET_CHANNEL", command)
	   gSetChannelDelayTimer:StartTimer(delay, delay_units)
	   gSetChannelDelayTimer._CallbackParam = command
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
	   return
    end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Additional Command Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
CMD_FUNC = {}

--[[
function CMD_FUNC.NUMBER_1 ()
  processDigits('1')
end

function CMD_FUNC.NUMBER_2 ()
  processDigits('2')
end

function CMD_FUNC.NUMBER_3 ()
  processDigits('3')
end

function CMD_FUNC.NUMBER_4 ()
  processDigits('4')
end

function CMD_FUNC.NUMBER_5 ()
  processDigits('5')
end

function CMD_FUNC.NUMBER_6 ()
  processDigits('6')
end

function CMD_FUNC.NUMBER_7 ()
  processDigits('7')
end

function CMD_FUNC.NUMBER_8 ()
  processDigits('8')
end

function CMD_FUNC.NUMBER_9 ()
  processDigits('9')
end

function CMD_FUNC.NUMBER_0 ()
  processDigits('0')
end

function CMD_FUNC.POUND ()
  processDigits('-')
end

function CMD_FUNC.ENTER()
  if g_ChannelBuffer ~= nil then
    local channel = g_ChannelBuffer
    g_ChannelBuffer = nil
    idBinding = 5001
    SET_CHANNEL(idBinding, channel)
    KillTimer (gChannelBufferTimer)
    
  else
	local command = CMDS ['ENTER']
	LogTrace("command = " .. command)
	PackAndQueueCommand('ENTER', command )
  end
end

function processDigits(digit)
	if g_ChannelBuffer==nil then
		if (digit~='-') then
			g_ChannelBuffer=digit
		end
	else
		g_ChannelBuffer=g_ChannelBuffer .. digit
	end
	KillTimer (gChannelBufferTimer)
	StartTimer (gChannelBufferTimer, tonumber(Properties['Channel Buffer Timer Value']), 'SECONDS', false)
end

function sendBufferedChannel ()
	local channel = g_ChannelBuffer
    g_ChannelBuffer = nil
	-- below needs to be changed so idBinding is used to select correct tuner type
	idBinding = 5001
	SET_CHANNEL(idBinding, channel)
	KillTimer (gChannelBufferTimer)
end
--]]

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	Helper Function: SEND_COMMAND_FROM_COMMAND_TABLE
	Parameters:
		idBinding: proxy id
		output: mod 1000 value of Output Connection id
		command_name: name of command to be sent
--]]
function SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, command_name)
    LogTrace("SEND_COMMAND_FROM_COMMAND_TABLE(), idBinding=" .. idBinding .. ", command_name=" .. command_name)

	-- This added to enable redirection of more commands to functions
	if (CMD_FUNC[command_name] ~= nil) and (type(CMD_FUNC[command_name])=='function') then
		CMD_FUNC[command_name] (command_name)
		return
    end

	-- TODO: create packet/command to send to the device
	local command = GetCommandFromCommandTable(idBinding, command_name)

	if (command == nil) then
		LogTrace("command is nil")
	else
		LogTrace("command = " .. command)
	end
	PackAndQueueCommand(command_name, command)
end

--[[
	Helper Function: GetCommandFromCommandTable
	Parameters:
		idBinding: proxy id
		output: mod 1000 value of Output Connection id
		command_name: name of command to be returned
--]]
function GetCommandFromCommandTable(idBinding, command_name)
    LogTrace("GetCommand()")
    local t = {}

    if (gControlMethod == "IR") then
	   t = CMDS_IR
    elseif (gControlMethod == "Serial") then
	   t = CMDS_Serial
    elseif (gControlMethod == "Network") then
	   t = CMDS_IP
    else
	   print(tostring(gControlMethod) .. " Control Method not handled!!!")
    end
	
    if (t[idBinding][command_name] ~= nil) then
	   return t[idBinding][command_name]
    elseif (t[command_name] ~= nil) then
	   return t[command_name]
    else
	   LogWarn('GetCommandFromCommandTable: command not defined - '.. command_name)
	   return nil
    end

end

--[[
	Helper Function: GetDeviceVolumeStatus
--]]
function GetCurrentChannel()
    LOG:Trace("GetCurrentChannel()")
    if (gControlMethod ~= "Network") then return end

	local command = CMDS_IP["GET_CURRENT_CHANNEL"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueueCommand("GET_CURRENT_CHANNEL", command)

end

function GetDeviceVolumeStatus()
    LOG:Trace("GetDeviceVolumeStatus()")
    if (gControlMethod ~= "Network") then return end

	-- TODO: verify table entry in Volume in QUERY table
	local command = CMDS_IP["GET_VOLUME_STATUS"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueueCommand("GET_VOLUME_STATUS", command)

	-- TODO: verify table entry in Mute in QUERY table
	command = CMDS_IP["GET_MUTE_STATUS"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueueCommand("GET_MUTE_STATUS", command)
end

function GetCurrentVolume()
    LOG:Trace("GetCurrentVolume()")
    if (gControlMethod ~= "Network") then return end

	-- TODO: verify table entry in Volume in QUERY table
	local command = CMDS_IP["GET_VOLUME_STATUS"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueueCommand("GET_VOLUME_STATUS", command)
end

function GetMuteState()
    LOG:Trace("GetDeviceVolumeStatus()")
    if (gControlMethod ~= "Network") then return end

	-- TODO: verify table entry in Mute in QUERY table
	command = CMDS_IP["GET_MUTE_STATUS"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueueCommand("GET_MUTE_STATUS", command)
end


function GetCurrentApp()
    LOG:Trace("GetCurrentApp()")
    if (gControlMethod ~= "Network") then return end

    local command = CMDS_IP["GET_CURRENT_APP"] or ""
    --LOG:Trace("command = " .. command)
    PackAndQueueCommand("GET_CURRENT_APP", command)
end

function GetMACAddressWired()
    LOG:Trace("GetMACAddressWired()")
    if (gControlMethod ~= "Network") then return end

    local command = CMDS_IP["GET_MACADDRESS_WIRED"] or ""
    --LOG:Trace("command = " .. command)
    PackAndQueueCommand("GET_MACADDRESS_WIRED", command)
end

function GetMACAddressWifi()
    LOG:Trace("GetMACAddressWifi()")
    if (gControlMethod ~= "Network") then return end

    local command = CMDS_IP["GET_MACADDRESS_WIFI"] or ""
    --LOG:Trace("command = " .. command)
    PackAndQueueCommand("GET_MACADDRESS_WIFI", command)
end

--[[
	Helper Function: GetDevicePowerStatus
--]]
function GetDevicePowerStatus()
    LOG:Trace("GetDevicePowerStatus()")
    if (gControlMethod ~= "Network") then return end

	-- TODO: verify table entry in Volume in QUERY table
	local command = CMDS_IP["GET_POWER_STATUS"] or ""
	LOG:Trace("command = " .. command)
	PackAndQueuePriority1Command("GET_POWER_STATUS", command)
end
 end)
package.preload['properties'] = (function (...)
--[[=============================================================================
    Properties Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

function ON_PROPERTY_CHANGED.NetworkKeepAliveIntervalSeconds(propertyValue)
	gNetworkKeepAliveInterval = tonumber(propertyValue)
end

function ON_PROPERTY_CHANGED.Password(propertyValue)
    gKey = PBKDF2(Properties["Password"])
    GetMACAddressWired()
    GetMACAddressWifi()
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)

    Description:
    Sets the value of the given property in the driver

    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed

    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)

	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end

 end)
package.preload['projector_init'] = (function (...)
--[[=============================================================================
    TV Protocol Initialization Functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end


PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_EARLY_INIT.projector_init()

end

function ON_DRIVER_INIT.projector_init()
	--LogTrace("ON_DRIVER_INIT.ProtocolDeclarations()")
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end
end

function ON_DRIVER_LATEINIT.projector_init()
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--index:  Proxy Command Name
		--value:  IR Code ID
		--Power
		["ON"]             = "101",
		["OFF"]            = "102",

		--Input Toggle
		["INPUT_TOGGLE"]="30",

		--Menu
		["INFO"] = "120",		--Display
		["GUIDE"] = "51008",	--Smart / Home	--"124"
		["MENU"] = "51026",		--MyApps
		["CANCEL"] = "121",		--Exit
		["UP"] = "113",
		["DOWN"] = "114",
		["LEFT"] = "115",
		["RIGHT"] = "116",
		["ENTER"] = "111",
		["RECALL"]         = "122", 	--Return (BACK)
		["PREV"]           = "122",	--Return (BACK)

		--Digits
		["NUMBER_0"]       = "10",
		["NUMBER_1"]       = "11",
		["NUMBER_2"]       = "12",
		["NUMBER_3"]       = "13",
		["NUMBER_4"]       = "14",
		["NUMBER_5"]       = "15",
		["NUMBER_6"]       = "16",
		["NUMBER_7"]       = "17",
		["NUMBER_8"]       = "18",
		["NUMBER_9"]       = "19",
		["STAR"]           = "51012",	--App / *	--"22"
		--["POUND"] 	    = "51003", --"24",	--List /-  --Key Code 53

		--Volume
		["MUTE_ON"]        = "",
		["MUTE_OFF"]       = "",
		["MUTE_TOGGLE"]    = "139",
		["PULSE_VOL_DOWN"] = "107",
		["PULSE_VOL_UP"]   = "106",
		["SET_VOLUME_LEVEL"] = "",

		--Tuning
		["SET_CHANNEL"]="",
		["PULSE_CH_UP"]="104",
		["START_CH_UP"]="",
		["SKIP_FWD"]="",		--used for PULSE_CH_UP in some Navigators
		["SCAN_FWD"]="",		--used for PULSE_CH_UP in some Navigators
		["PULSE_CH_DOWN"]="105",
		["START_CH_DOWN"]="",
		["SKIP_REV"]="",		--used for PULSE_CH_DOWN in some Navigators
		["SCAN_REV"]="",		--used for PULSE_CH_DOWN in some Navigators
		
		--Transport
		["PLAY"]="1",
		["STOP"]="2",
		["PAUSE"]="3",
		["SKIP_FWD"]="4",
		["SCAN_FWD"]="6",
		["SKIP_REV"]="5",
		["SCAN_REV"]="7",
		["RECORD"]="8",
		
		['CLOSED_CAPTIONED'] = '51013',
		['PROGRAM_A']		= '51007',  --RED
		['PROGRAM_B']		= '51006',  --GREEN
		['PROGRAM_C']		= '51005',  --YELLOW
		['PROGRAM_D']		= '51004',  --BLUE
		
-- Additional Device specific commands
		['Aspect Ratio'] 		= '51017',
		['AV Mode'] 			= '51018',
		['SIMPLINK'] 			= '51022',
		['3D'] 				= '51024',
		['Flashback'] 			= '51020',
		['Favorite Channel'] 	= '51019',
		['User Guide'] 		= '51025',
		['Program Guide'] 		= '51021',
		['Soccer'] 			= '51023',
		['Settings'] 			= '51027',
	}
	
	if (Properties["Tuner Region"] == "Europe, Mid-East, Colombia, Asia except South Korea and Japan Model") then
	   CMDS_IR["POUND"] = "51003" 	--"24",	--List /- 			--Key Code 53
	   CMDS_IR["HYPHEN"] = "51003"
     else
	   CMDS_IR["POUND"] = "51002" 	--"24",	--List, - (ATSC Only) 	--Key Code 4C
	   CMDS_IR["HYPHEN"] = "51002"
	end


	CMDS_IR[PROJECTOR_PROXY_BINDINGID] = {}

	CMDS_IR[AMAZON_PROXY_BINDINGID] = {
		--these entries are only required if
		--specific command strings are required for the Media Player
	--[[
		--Transport
		["PLAY"]="",
		["STOP"]="",
		["PAUSE"]="",
		["SKIP_FWD"]="",
		["SCAN_FWD"]="",
		["SKIP_REV"]="",
		["SCAN_REV"]="",

		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",
		["RECALL"]         = "",
		["PREV"]           = "",

		--Keypad
		["STAR"]="TAC",
		["NUMBER_0"]="",
		["NUMBER_1"]="",
		["NUMBER_2"]="",
		["NUMBER_3"]="",
		["NUMBER_4"]="",
		["NUMBER_5"]="",
		["NUMBER_6"]="",
		["NUMBER_7"]="",
		["NUMBER_8"]="",
		["NUMBER_9"]="",
	--]]
	}
	CMDS_IR[CHANNELPLUS_PROXY_BINDINGID] = {}
	CMDS_IR[LGCONTENTSTORE_PROXY_BINDINGID] = {}
	CMDS_IR[NETFLIX_PROXY_BINDINGID] = {}
	CMDS_IR[VUDU_PROXY_BINDINGID] = {}
	CMDS_IR[WEBBROWSER_PROXY_BINDINGID] = {}
	CMDS_IR[YOUTUBE_PROXY_BINDINGID] = {}
	CMDS_IR[GOOGLEPLAY_PROXY_BINDINGID] = {}
	CMDS_IR[HULU_PROXY_BINDINGID] = {}
	CMDS_IR[SLINGTV_PROXY_BINDINGID] = {}
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()")

	CMDS_Serial = {
		--index:  Proxy Command Name
		--value:  Protocol Command Data

		--Power
		["ON"]             = 'ka SET_ID 01',
		["OFF"]            = "ka SET_ID 00",

		--Input Toggle
		["INPUT_TOGGLE"]	= "",

		--Menu
		["INFO"]			= "mc SET_ID AA",
		["GUIDE"]			= "mc SET_ID 7C",	--Smart / Home
		["MENU"]			= "mc SET_ID 43",
		["CANCEL"]		= "mc SET_ID 5B",	
		["UP"]			= "mc SET_ID 40",
		["DOWN"]			= "mc SET_ID 41",
		["LEFT"]			= "mc SET_ID 07",
		["RIGHT"]			= "mc SET_ID 06",
		["ENTER"] 		= "mc SET_ID 44",	-- For LG will be completed by function
		["RECALL"]		= "mc SET_ID 28",
		["PREV"]			= "mc SET_ID 28",

		--Digits
		["NUMBER_0"]       = "mc SET_ID 10",
		["NUMBER_1"]       = "mc SET_ID 11",
		["NUMBER_2"]       = "mc SET_ID 12",
		["NUMBER_3"]       = "mc SET_ID 13",
		["NUMBER_4"]       = "mc SET_ID 14",
		["NUMBER_5"]       = "mc SET_ID 15",
		["NUMBER_6"]       = "mc SET_ID 16",
		["NUMBER_7"]       = "mc SET_ID 17",
		["NUMBER_8"]       = "mc SET_ID 18",
		["NUMBER_9"]       = "mc SET_ID 19",
		["STAR"]           = "mc SET_ID 9F", 
		--["POUND"] 	    = "mc SET_ID 53", 	--"24",	--List /- 
		
		--Volume
		["MUTE_ON"]        = "ke SET_ID 00",
		["MUTE_OFF"]       = "ke SET_ID 01",
		["MUTE_TOGGLE"]    = "mc SET_ID 09",
		["PULSE_VOL_DOWN"] = "mc SET_ID 03",
		["PULSE_VOL_UP"]   = "mc SET_ID 02",
		["SET_VOLUME_LEVEL"] = "kf SET_ID ", -- Completed by SET_VOLUME function

		--Tuning
		["SET_CHANNEL"]		= "ma SET_ID ", -- Completed by Function
		["PULSE_CH_UP"]		= "mc SET_ID 00",
		["START_CH_UP"]		= "mc SET_ID 00",
		--["SKIP_FWD"]		= "",		--used for PULSE_CH_UP in some Navigators
		--["SCAN_FWD"]		= "",		--used for PULSE_CH_UP in some Navigators
		["PULSE_CH_DOWN"]	= "mc SET_ID 01",
		["START_CH_DOWN"]	= "mc SET_ID 01",
		--["SKIP_REV"]		= "",		--used for PULSE_CH_DOWN in some Navigators
		--["SCAN_REV"]		= "",		--used for PULSE_CH_DOWN in some Navigators

		--Status Query
		["GET_VOLUME_STATUS"]="",
		["GET_MUTE_STATUS"]="",
		["GET_POWER_STATUS"]="",

		['CLOSED_CAPTIONED'] = 'mc SET_ID 39',
		['PROGRAM_A']		= 'mc SET_ID 72',  --RED
		['PROGRAM_B']		= 'mc SET_ID 71',  --GREEN
		['PROGRAM_C']		= 'mc SET_ID 63',  --YELLOW
		['PROGRAM_D']		= 'mc SET_ID 61',  --BLUE

-- As TV can have it's own DVR then these need to be defined here
		['PLAY'] 			= "mc SET_ID B0",
		['PAUSE'] 		= "mc SET_ID BA",
		['STOP'] 			= "mc SET_ID B1",
		['SCAN_FWD'] 		= "mc SET_ID 8E",
		['SCAN_REV'] 		= "mc SET_ID 8F",
		['SKIP_FWD'] 		= "mc SET_ID 8E",
		['SKIP_REV'] 		= "mc SET_ID 8F",
		['RECORD'] 		= "mc SET_ID BD",
		['PVR'] 			= 'mc SET_ID 7C', -- Home/Menu

-- Additional Device specific commands
		['Aspect Ratio'] 		= 'mc SET_ID 79',
		['AV Mode'] 			= 'mc SET_ID 30',
		['SIMPLINK'] 			= 'mc SET_ID 7E',
		['3D'] 				= 'mc SET_ID DC',
		['Flashback'] 			= 'mc SET_ID 1A',
		['Favorite Channel'] 	= 'mc SET_ID 1E',
		['User Guide'] 		= 'mc SET_ID 7A',
		['Program Guide'] 		= 'mc SET_ID AB',
		['Soccer'] 			= 'mc SET_ID BB',
		
	}
	

	if (Properties["Tuner Region"] == "Europe, Mid-East, Colombia, Asia except South Korea and Japan Model") then
	   CMDS_Serial["POUND"] = "mc SET_ID 53" 	--"24",	--List /- 
	   CMDS_Serial["HYPHEN"] = "mc SET_ID 53"
     else
	   CMDS_Serial["POUND"] = "mc SET_ID 4C" 	--"24",	--List, - (ATSC Only) 
	   CMDS_Serial["HYPHEN"] = "mc SET_ID 4C"
	end	

	CMDS_Serial[PROJECTOR_PROXY_BINDINGID] = {}

	CMDS_Serial[AMAZON_PROXY_BINDINGID] = {}
	CMDS_Serial[CHANNELPLUS_PROXY_BINDINGID] = {}
	CMDS_Serial[LGCONTENTSTORE_PROXY_BINDINGID] = {}
	CMDS_Serial[NETFLIX_PROXY_BINDINGID] = {}
	CMDS_Serial[VUDU_PROXY_BINDINGID] = {}
	CMDS_Serial[WEBBROWSER_PROXY_BINDINGID] = {}
	CMDS_Serial[YOUTUBE_PROXY_BINDINGID] = {}
	CMDS_Serial[GOOGLEPLAY_PROXY_BINDINGID] = {}
	CMDS_Serial[HULU_PROXY_BINDINGID] = {}
	CMDS_Serial[SLINGTV_PROXY_BINDINGID] = {}
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_IP()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IP()")

	CMDS_IP = {
		--index:  Proxy Command Name
		--value:  Protocol Command Data

		--Power
		["ON"]             = '',	--WOL is used
		["OFF"]            = "POWER off",

		--Input Toggle
		["INPUT_TOGGLE"]	= "KEY_ACTION deviceinput",

		--Menu
		["INFO"]			= "KEY_ACTION programminfo",
		["GUIDE"]			= "KEY_ACTION programguide",	--programguide	--Smart / Home
		["MENU"]			= "KEY_ACTION smarthome",	--KEY_ACTION programguide, KEY_ACTION settingmenu
		["CANCEL"]		= "KEY_ACTION returnback",	
		["UP"]			= "KEY_ACTION arrowup",
		["DOWN"]			= "KEY_ACTION arrowdown",
		["LEFT"]			= "KEY_ACTION arrowleft",
		["RIGHT"]			= "KEY_ACTION arrowright",
		["ENTER"] 		= "KEY_ACTION ok",	-- For LG will be completed by function
		["RECALL"]		= "KEY_ACTION previouschannel",
		["PREV"]			= "KEY_ACTION previouschannel",

		--Digits
		["NUMBER_0"]       = "KEY_ACTION number0",
		["NUMBER_1"]       = "KEY_ACTION number1",
		["NUMBER_2"]       = "KEY_ACTION number2",
		["NUMBER_3"]       = "KEY_ACTION number3",
		["NUMBER_4"]       = "KEY_ACTION number4",
		["NUMBER_5"]       = "KEY_ACTION number5",
		["NUMBER_6"]       = "KEY_ACTION number6",
		["NUMBER_7"]       = "KEY_ACTION number7",
		["NUMBER_8"]       = "KEY_ACTION number8",
		["NUMBER_9"]       = "KEY_ACTION number9",
		["STAR"]           = "", 
		--["POUND"] 	    = "KEY_ACTION channellist", 	--"24",	--List /- 
		
		--Volume
		["MUTE_ON"]        = "VOLUME_MUTE on",
		["MUTE_OFF"]       = "VOLUME_MUTE off",
		["MUTE_TOGGLE"]    = "KEY_ACTION volumemute",
		["PULSE_VOL_DOWN"] = "KEY_ACTION volumedown",
		["PULSE_VOL_UP"]   = "KEY_ACTION volumeup",
		["SET_VOLUME_LEVEL"] = "VOLUME_CONTROL ", -- Completed by SET_VOLUME function

		--Tuning
		["SET_CHANNEL"]		= "", -- Completed by Function
		["PULSE_CH_UP"]		= "KEY_ACTION channelup",
		["START_CH_UP"]		= "KEY_ACTION channelup",
		--["SKIP_FWD"]		= "",		--used for PULSE_CH_UP in some Navigators
		--["SCAN_FWD"]		= "",		--used for PULSE_CH_UP in some Navigators
		["PULSE_CH_DOWN"]	= "KEY_ACTION channeldown",
		["START_CH_DOWN"]	= "KEY_ACTION channeldown",
		--["SKIP_REV"]		= "",		--used for PULSE_CH_DOWN in some Navigators
		--["SCAN_REV"]		= "",		--used for PULSE_CH_DOWN in some Navigators

		--Status Query
		["GET_CURRENT_CHANNEL"]	=	"CURRENT_CH",
		["GET_VOLUME_STATUS"]	=	"CURRENT_VOL",
		["GET_MUTE_STATUS"]		=	"MUTE_STATE",
		["GET_POWER_STATUS"]	=	"GET_IPCONTROL_STATE",
		["GET_CURRENT_APP"]		=	"CURRENT_APP",
		["GET_MACADDRESS_WIRED"]	=	"GET_MACADDRESS wired",
		["GET_MACADDRESS_WIFI"]	=	"GET_MACADDRESS wifi",
		
    
		['CLOSED_CAPTIONED'] = 'KEY_ACTION captionsubtitle',
		['PROGRAM_A']		= 'KEY_ACTION redbutton',  --RED
		['PROGRAM_B']		= 'KEY_ACTION greenbutton',  --GREEN
		['PROGRAM_C']		= 'KEY_ACTION yellowbutton',  --YELLOW
		['PROGRAM_D']		= 'KEY_ACTION bluebutton',  --BLUE

-- As TV can have it's own DVR then these need to be defined here
		['PLAY'] 			= "KEY_ACTION play",
		['PAUSE'] 			= "KEY_ACTION ok",
		['STOP'] 			= "",
		['SCAN_FWD'] 		= "KEY_ACTION fastforward",
		['SCAN_REV'] 		= "KEY_ACTION rewind",
		['SKIP_FWD'] 		= "",
		['SKIP_REV'] 		= "",
		['RECORD'] 			= "KEY_ACTION record",
		['PVR'] 			= "", -- Home/Menu

-- Additional Device specific commands
		['Aspect Ratio'] 		= "KEY_ACTION aspectratio",
		['AV Mode'] 			= "KEY_ACTION avmode",
		['SIMPLINK'] 			= "KEY_ACTION simplelink",
		['3D'] 				= "KEY_ACTION 3d",
		['Flashback'] 			= "KEY_ACTION returnback",
		['Favorite Channel'] 	= "KEY_ACTION favoritechannel",
		['User Guide'] 		= "KEY_ACTION userguide",
		['Program Guide'] 		= "KEY_ACTION programguide",
		['Soccer'] 			= "KEY_ACTION soccerscreen",
		["Settings"]			= "KEY_ACTION quickmenu",	--settingmenu  
		
	}
	
	if (Properties["Tuner Region"] == "Europe, Mid-East, Colombia, Asia except South Korea and Japan Model") then
	   CMDS_IP["POUND"] = "KEY_ACTION channellist" 	--"24",	--List /- 
	   CMDS_IP["HYPHEN"] = "KEY_ACTION channellist"
     else
	   CMDS_IP["POUND"] = "KEY_ACTION channellist" 	--"24",	--List, - (ATSC Only) 
	   CMDS_IP["HYPHEN"] = "KEY_ACTION channellist"
	end	

	
	CMDS_IP[PROJECTOR_PROXY_BINDINGID] = {}

	CMDS_IP[AMAZON_PROXY_BINDINGID] = {}
	CMDS_IP[CHANNELPLUS_PROXY_BINDINGID] = {}
	CMDS_IP[LGCONTENTSTORE_PROXY_BINDINGID] = {}
	CMDS_IP[NETFLIX_PROXY_BINDINGID] = {}
	CMDS_IP[VUDU_PROXY_BINDINGID] = {}
	CMDS_IP[WEBBROWSER_PROXY_BINDINGID] = {}
	CMDS_IP[YOUTUBE_PROXY_BINDINGID] = {}
	CMDS_IP[GOOGLEPLAY_PROXY_BINDINGID] = {}
	CMDS_IP[HULU_PROXY_BINDINGID] = {}
	CMDS_IP[SLINGTV_PROXY_BINDINGID] = {}

end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	LogTrace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	tInputCommandMap_IR = {
		['AV']='19130',
		['AV 1']='19130',
		['AV 2']='19131',
		
		['COMPONENT']='19097',
		['COMPONENT 1']='19097',
		['COMPONENT 2']='19098',
		['COMPONENT 3']='50998',
		
		['RGB-PC']='19121',
		
		['HDMI 1']='28425',
		['HDMI 2']='28426',
		['HDMI 3']='41166',
		['HDMI 4']='50997',
		
		['Antenna TV (ATV)']='19093',				--EMEA, AMERICAS
		['Cable TV (CATV)']='34960',				--EMEA, AMERICAS
		['Antenna TV (DTV)']='19093',				--EMEA, AMERICAS, JAPAN
	     ['Cable TV (CADTV)']='34960',				--EMEA, AMERICAS
		
		['Satellite TV (SDTV)']='19093',				--EMEA
		
		['BS (Broadcasting Satellite)']='19093',		--JAPAN
		['CS1 (Communication Satellite 1)']='190933',	--JAPAN
		['CS2 (Communication Satellite 2)']='19093',	--JAPAN
	}
	
	tInputCommandMap_Serial = {
		--index:  Connection Name
		--value:  Protocol Command Data
		['AV']='xb SET_ID 20',
		['AV 1']='xb SET_ID 20',
		['AV 2']='xb SET_ID 21',
		
		['COMPONENT']='xb SET_ID 40',
		['COMPONENT 1']='xb SET_ID 40',
		['COMPONENT 2']='xb SET_ID 41',
		['COMPONENT 3']='xb SET_ID 42',
		
		['RGB-PC']='xb SET_ID 60',
		
		['HDMI 1']='xb SET_ID 90',
		['HDMI 2']='xb SET_ID 91',
		['HDMI 3']='xb SET_ID 92',
		['HDMI 4']='xb SET_ID 93',
		
		['Antenna TV (ATV)']='xb SET_ID 10',				--EMEA, AMERICAS
		['Cable TV (CATV)']='xb SET_ID 11',				--EMEA, AMERICAS
		['Antenna TV (DTV)']='xb SET_ID 00',				--EMEA, AMERICAS, JAPAN
	     ['Cable TV (CADTV)']='xb SET_ID 01',				--EMEA, AMERICAS
		
		['Satellite TV (SDTV)']='xb SET_ID 02',				--EMEA
		
		['BS (Broadcasting Satellite)']='xb SET_ID 02',		--JAPAN
		['CS1 (Communication Satellite 1)']='xb SET_ID 03',	--JAPAN
		['CS2 (Communication Satellite 2)']='xb SET_ID 04',	--JAPAN
	}
	

	tInputCommandMap_IP = {
		--index:  Connection Name
		--value:  Protocol Command Data
		['AV']='APP_LAUNCH com.webos.app.externalinput.av1',
		['AV 1']='APP_LAUNCH com.webos.app.externalinput.av1',
		['AV 2']='',
		
		['COMPONENT']='APP_LAUNCH com.webos.app.externalinput.component',
		['COMPONENT 1']='APP_LAUNCH com.webos.app.externalinput.component',
		['COMPONENT 2']='',
		['COMPONENT 3']='',
		
		['RGB-PC']='',
		
		['HDMI 1']='APP_LAUNCH com.webos.app.hdmi1',
		['HDMI 2']='APP_LAUNCH com.webos.app.hdmi2',
		['HDMI 3']='APP_LAUNCH com.webos.app.hdmi3',
		['HDMI 4']='APP_LAUNCH com.webos.app.hdmi4',

		['Antenna TV (ATV)']='APP_LAUNCH com.webos.app.livetv',	--atv, livetv			--EMEA, AMERICAS
		['Cable TV (CATV)']='APP_LAUNCH com.webos.app.livetv',		--catv, livetv			--EMEA, AMERICAS
		['Antenna TV (DTV)']='APP_LAUNCH com.webos.app.livetv',	--dtv, livetv			--EMEA, AMERICAS, JAPAN
	     ['Cable TV (CADTV)']='APP_LAUNCH com.webos.app.livetv',	--cadtv, livetv		--EMEA, AMERICAS
		
		--IP Protocol is only for USA models
		['Satellite TV (SDTV)']='',			--EMEA
		
		['BS (Broadcasting Satellite)']='',	--JAPAN
		['CS1 (Communication Satellite 1)']='',	--JAPAN
		['CS2 (Communication Satellite 2)']='',	--JAPAN
		
		--apps, also included in gAppLaunchCommandMap
		['Netflix']='APP_LAUNCH netflix',
		['Amazon']='APP_LAUNCH amazon',
		['VUDU']='APP_LAUNCH vudu',
		['Hulu']='APP_LAUNCH com.hulu.hulu',
		['HuluPLUS']='APP_LAUNCH hulu-simplesmart',
		['Channel Plus']='APP_LAUNCH channelplus.us',
		['YouTube']='APP_LAUNCH youtube.leanback.v4',
		['Google Play']='APP_LAUNCH googleplaymovieswebos',
		['Web Browser']='APP_LAUNCH com.webos.app.browser',
		['Sling TV']='APP_LAUNCH com.movenetworks.app.sling-tv-sling-production',
		['LG Content Store']='APP_LAUNCH com.webos.app.discovery',
		['Gallery']='APP_LAUNCH com.webos.app.igallery',
		
	}
	
--[[
	tInputCommandMap_IP = {
		--index:  Connection Name
		--value:  Protocol Command Data
		['AV']='INPUT_SELECT avav1',
		['AV 1']='INPUT_SELECT avav1',
		['AV 2']='',
		
		['COMPONENT']='INPUT_SELECT component1',
		['COMPONENT 1']='INPUT_SELECT component1',
		['COMPONENT 2']='',
		['COMPONENT 3']='',
		
		['RGB-PC']='',
		
		['HDMI 1']='INPUT_SELECT hdmi1',
		['HDMI 2']='INPUT_SELECT hdmi2',
		['HDMI 3']='INPUT_SELECT hdmi3',
		['HDMI 4']='',
		
		['Antenna TV (ATV)']='INPUT_SELECT atv',				--EMEA, AMERICAS
		['Cable TV (CATV)']='INPUT_SELECT catv',				--EMEA, AMERICAS
		['Antenna TV (DTV)']='INPUT_SELECT dtv',				--EMEA, AMERICAS, JAPAN
	     ['Cable TV (CADTV)']='INPUT_SELECT cadtv',				--EMEA, AMERICAS
		
		--IP Protocol is only for USA models
		['Satellite TV (SDTV)']='',			--EMEA
		
		['BS (Broadcasting Satellite)']='',	--JAPAN
		['CS1 (Communication Satellite 1)']='',	--JAPAN
		['CS2 (Communication Satellite 2)']='',	--JAPAN
	}
--]]

	--tInputResponseMap = ReverseTable(tInputCommandMap)	-- Reverses the tInputCommandMap table
	tInputResponseMap = {
		['20']='AV 1',
		['21']='AV 2',
		
		['40']='COMPONENT',
		['41']='COMPONENT 2',
		['42']='COMPONENT 3',

		['60']='RGB-PC',
		
		['90']='HDMI 1',
		['91']='HDMI 2',
		['92']='HDMI 3',
		['93']='HDMI 4',
    
		['10']='Antenna TV (ATV)',				--EMEA, AMERICAS
		['11']='Cable TV (CATV)',				--EMEA, AMERICAS
		['00']='Antenna TV (DTV)',				--EMEA, AMERICAS, JAPAN
	     ['01']='Cable TV (CADTV)',				--EMEA, AMERICAS
		
		['02']='Satellite TV (SDTV)',				--EMEA
		
		--['02']='BS (Broadcasting Satellite)',		--JAPAN
		['03']='CS1 (Communication Satellite 1)',	--JAPAN
		['04']='CS2 (Communication Satellite 2)',	--JAPAN
	}
	
	tInputResponseMap_IP = {
		['com.webos.app.externalinput.av1']='AV 1',
		--['21']='AV 2',
		
		['com.webos.app.externalinput.component']='COMPONENT',
		--['41']='COMPONENT 2',
		--['42']='COMPONENT 3',

		--['60']='RGB-PC',
		
		['com.webos.app.hdmi1']='HDMI 1',
		['com.webos.app.hdmi2']='HDMI 2',
		['com.webos.app.hdmi3']='HDMI 3',
		['com.webos.app.hdmi4']='HDMI 4',
    
		['com.webos.app.livetv']='Antenna TV (DTV)',
		--['atv']='Antenna TV (ATV)',				--EMEA, AMERICAS
		--['catv']='Cable TV (CATV)',				--EMEA, AMERICAS
		--['dtv']='Antenna TV (DTV)',				--EMEA, AMERICAS, JAPAN
	     --['cadtv']='Cable TV (CADTV)',				--EMEA, AMERICAS
		
		--['02']='Satellite TV (SDTV)',				--EMEA
		
		--['02']='BS (Broadcasting Satellite)',		--JAPAN
		--['03']='CS1 (Communication Satellite 1)',	--JAPAN
		--['04']='CS2 (Communication Satellite 2)',	--JAPAN
	}

	----------------------------------------- [*I/O HELPER TABLES*] -----------------------------------------
	tOutputConnMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Output Connection Name
		[0] = 'Output',
	}

	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		--[0] = {Name = "INPUT HDMI 1",BindingID = PROJECTOR_PROXY_BINDINGID,},
		--The first two entries below are for the antenna connections which are bound to the tuner proxy; the in values will need to be updated base on the connection ID.
		[1] = {Name = 'AV 1',BindingID = 5001,},
		[2] = {Name = 'AV 2',BindingID = 5001,},
		[11] = {Name = 'COMPONENT',BindingID = 5001,},
		[12] = {Name = 'COMPONENT 2',BindingID = 5001,},
		[13] = {Name = 'COMPONENT 3',BindingID = 5001,},
		[21] = {Name = 'HDMI 1',BindingID = 5001,},
		[22] = {Name = 'HDMI 2',BindingID = 5001,},
		[23] = {Name = 'HDMI 3',BindingID = 5001,},
		[24] = {Name = 'HDMI 4',BindingID = 5001,},
		[31] = {Name = 'RGB-PC',BindingID = 5001,},
		[101] = {Name = 'Antenna TV (ATV)',BindingID = 5001,},
		[102] = {Name = 'Cable TV (CATV)',BindingID = 5001,},
		[103] = {Name = 'Antenna TV (DTV)',BindingID = 5001,},
		[104] = {Name = 'Cable TV (CADTV)',BindingID = 5001,},
		[105] = {Name = 'Satellite TV (SDTV)',BindingID = 5001,},
		[106] = {Name = 'Antenna Radio (Radio)',BindingID = 5001,},
		[107] = {Name = 'Satellite Radio (S-Radio)',BindingID = 5001,},
		[108] = {Name = 'Cable Radio (CA-Radio)',BindingID = 5001,},
		[109] = {Name = 'BS (Broadcasting Satellite)',BindingID = 5001,},
		[110] = {Name = 'CS1 (Communication Satellite 1)',BindingID = 5001,},
		[111] = {Name = 'CS2 (Communication Satellite 2)',BindingID = 5001,},
		
		--[[
		[201] = {Name = 'Amazon',BindingID = 5001,},
		[202] = {Name = 'Channel Plus',BindingID = 5001,},
		[203] = {Name = 'LG Content Store',BindingID = 5001,},
		[204] = {Name = 'Netflix',BindingID = 5001,},
		[205] = {Name = 'VUDU',BindingID = 5001,},
		[206] = {Name = 'Web Browser',BindingID = 5001,},
		[207] = {Name = 'YouTube',BindingID = 5001,},
		[208] = {Name = 'Google Play',BindingID = 5001,},
		[209] = {Name = 'Hulu',BindingID = 5001,},
		[210] = {Name = 'Sling TV',BindingID = 5001,},
		--]]
		
	}

	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--["INPUT HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		['AV 1'] = {ID = 1,BindingID = 5001,},
		['AV 2'] = {ID = 2,BindingID = 5001,},
		['COMPONENT'] = {ID = 11,BindingID = 5001,},
		['COMPONENT 2'] = {ID = 12,BindingID = 5001,},
		['COMPONENT 3'] = {ID = 13,BindingID = 5001,},
		['HDMI 1'] = {ID = 21,BindingID = 5001,},
		['HDMI 2'] = {ID = 22,BindingID = 5001,},
		['HDMI 3'] = {ID = 23,BindingID = 5001,},
		['HDMI 4'] = {ID = 24,BindingID = 5001,},
		['RGB-PC'] = {ID = 31,BindingID = 5001,},
		['Antenna TV (ATV)'] = {ID = 101,BindingID = 5001,},
		['Cable TV (CATV)'] = {ID = 102,BindingID = 5001,},
		['Antenna TV (DTV)'] = {ID = 103,BindingID = 5001,},
		['Cable TV (CADTV)'] = {ID = 104,BindingID = 5001,},
		['Satellite TV (SDTV)'] = {ID = 105,BindingID = 5001,},
		['Antenna Radio (Radio)'] = {ID = 106,BindingID = 5001,},
		['Satellite Radio (S-Radio)'] = {ID = 107,BindingID = 5001,},
		['Cable Radio (CA-Radio)'] = {ID = 108,BindingID = 5001,},
		['BS (Broadcasting Satellite)'] = {ID = 109,BindingID = 5001,},
		['CS1 (Communication Satellite 1)'] = {ID = 110,BindingID = 5001,},
		['CS2 (Communication Satellite 2)'] = {ID = 111,BindingID = 5001,},
		
		--[[
		['Amazon'] = {ID = 201,BindingID = 5001,},
		['Channel Plus'] = {ID = 202,BindingID = 5001,},
		['LG Content Store'] = {ID = 203,BindingID = 5001,},
		['Netflix'] = {ID = 204,BindingID = 5001,},
		['VUDU'] = {ID = 204,BindingID = 5001,},
		['Web Browser'] = {ID = 206,BindingID = 5001,},
		['YouTube'] = {ID = 207,BindingID = 5001,},
		['Google Play'] = {ID = 208,BindingID = 5001,},
		['Hulu'] = {ID = 209,BindingID = 5001,},
		['Sling TV'] = {ID = 210,BindingID = 5001,},
		--]]
	}

end

-- What goes here - not in documentation
function PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()")

end

function PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()")

	--[[
	TODO: Update tVolumeCurve table based upon devices volume range and optimal ramping steps
		tVolumeCurve is used to plot values that will be used
		to build volume commands during volume ramping.
		Specifically, they are used in GetNextVolumeCurveValue() which is
		called from the ContinueVolumeRamping() function.
		The values entered here should correlate to the volume parameters as defined in the device protocol.
		Below is an example, the table for your device may use different values.
		Also the number and size of steps may differ for your device. Adjust values to get
		a smooth volume ramping from low to high volume.
		NOTE: values must be in decimal (i.e. not hexadecimal...)
	--]]
	tVolumeCurve = {
		-- 5 unit STEPS [0 to 20]
		0,5,10,15,20,
		-- 4 unit STEPS [24 to 58]
		24,28,32,36,38,
		42,46,49,52,55,58,
		-- 2 unit STEPS [60 to 90]
		60,62,64,66,68,
		70,72,74,76,78,
		80,82,84,86,88,90,
		-- 1 unit STEPS [90 to 100]
		91,92,93,94,95,
		96,97,98,99,100,
	}

end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

 end)
package.preload['device_messages'] = (function (...)
--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    GetMessage()

    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.

    Parameters:
    None

    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
    local message, pos
	
    if (gReceiveBuffer:len() > 0) then
	   if (gControlMethod == "Network") then
			message = lg_decrypt(gReceiveBuffer)
			gReceiveBuffer = ""
	   else
		  local pattern = "(%a.-)x()"
		  message, pos = string.match(gReceiveBuffer, pattern)
		  if (message == nil) then
			 --LOG:Info("Do not have a complete message")
			 return ""
		  end
		  gReceiveBuffer = gReceiveBuffer:sub(pos)
	   end
	end

	return message

end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
gSetInputRetryCounter = 0
gSetChannelRetryCounter = 0

function HandleMessage(message)
    LogTrace(string.format("HandleMessage. Message is ==>%s<==", message))
    
    local name, value
    if (gControlMethod == "Network") then  
	   
	   local cmd = gDevMsgCallback
	   if (cmd == "" or cmd == nil) then LogTrace("HandleMessage: gDevMsgCallback is nil") return end
	   
	   print("HandleMessage: cmd = " .. cmd .. ", message = " .. message)
	   
	   if (message == "NG") then 
		  LogWarn("Command '" .. cmd .. "' returned and error!") 
		  if (cmd == "SET_INPUT") then
			 gSetInputRetryCounter = gSetInputRetryCounter or 0
			 gSetInputRetryCounter = gSetInputRetryCounter + 1
			 if gSetInputRetryCounter < 4 then
				gSetInputDelayTimer:StartTimer(1, "SECONDS")
			 else
				gSetInputRetryCounter = 0
				LogWarn("SET_INPUT failed 3 Retries, Exiting Retry.")
			 end
		  end
		  if (cmd == "SET_CHANNEL") then
			 gSetChannelRetryCounter = gSetChannelRetryCounter or 0
			 gSetChannelRetryCounter = gSetChannelRetryCounter + 1
			 if gSetChannelRetryCounter < 4 then
				gSetChannelDelayTimer:StartTimer(1, "SECONDS")
			 else
				gSetChannelRetryCounter = 0
				LogWarn("SET_CHANNEL failed 3 Retries, Exiting Retry.")
			 end
		  end
		  
		  --NACK, no further processing necessary...
		  gCon:HandleACK()
		  return 
	   elseif (message == "OK") then 
		  if (cmd == "OFF") then
			 gQueryPowerDelayTimer:StartTimer(1, "SECONDS")
		  elseif (cmd == "SET_INPUT") then
			 gQueryInputDelayTimer:StartTimer(4, "SECONDS")	--2 seconds too short during power...
		  elseif (cmd == "MUTE_OFF" or cmd == "MUTE_ON" or cmd == "MUTE_TOGGLE") then
			 gQueryMuteDelayTimer:StartTimer(1, "SECONDS")
		  elseif (cmd == "SET_VOLUME_LEVEL" or cmd == "PULSE_VOL_DOWN" or cmd == "PULSE_VOL_UP") then
			 GetCurrentVolume()
		  elseif (cmd == "SET_CHANNEL" or cmd == "PULSE_CH_DOWN" or cmd == "PULSE_CH_UP") then
			 gQueryChannelDelayTimer:StartTimer(3, "SECONDS")
		  end
		  
		  --ACK, no further processing necessary...
		  gCon:HandleACK()
		  return
	   else
		  name = cmd
		  value = message
	   end
    else
	   --unit ACKS 'BU' when turned off...not sure why but we do not need this data...
        message = string.gsub(message, "BU", "")
	   local cmd, ack, data, set_id
	   local buflen=string.len(message)

	   cmd = string.sub(message, 1, 1)
	   set_id = string.sub(message, 3, 4)
	   ack = string.sub(message, 6, 7)
	   data = string.sub(message, 8, 9)
	   
	   if (ack == "NG") then LogWarn("Command '" .. cmd .. "' returned and error!") return end

	   if (cmd =='a') and (buflen == 13) then -- EMEA channel change
		  name = 'CHANNEL_EMEA'
		  value = message
	   elseif (cmd =='a') and (buflen == 19) then -- US channel change
		  name = 'CHANNEL_US'
		  value = message
	   elseif (cmd == 'a') then
		  name = 'POWER'
		  value = data
	   elseif (cmd == 'b') then
		  name = 'INPUT'
		  value = data
	   elseif (cmd == 'c') then
		  -- remote Key Codes...
		  gCon:HandleACK()
		  return
	   elseif (cmd == 'e') then
		  name = 'MUTE'
		  value = data
	   elseif (cmd == 'f') then
		    name = 'VOLUME'
		    value = data
	   end

	   if (name == nil) then return end
	   if (value == nil or value == "") then return end
	   
    end
    
    DispatchMessage(name, value)

    gCon:HandleACK()
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LogInfo("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhanded command = " .. MsgKey)
	end
end


--===========================NETWORK CONTROL===================================--
function DEV_MSG.GET_MACADDRESS_WIRED(value)
    LogTrace("DEV_MSG.GET_MACADDRESS_WIRED(), value = " .. value)
	
    local mac = value or ""
    if (mac:find("%x%x:%x%x:%x%x:%x%x:%x%x:%x%x")) then
	   UpdateProperty("Wired MAC Address", mac)
    end
end

function DEV_MSG.GET_MACADDRESS_WIFI(value)
    LogTrace("DEV_MSG.GET_MACADDRESS_WIFI(), value = " .. value)
    
    local mac = value or ""
    if (mac:find("%x%x:%x%x:%x%x:%x%x:%x%x:%x%x")) then
	   UpdateProperty("Wi-Fi MAC Address", value)
    end
end

function DEV_MSG.GET_POWER_STATUS(value)
	LogTrace("DEV_MSG.GET_POWER_STATUS(), value = " .. value)

    	if gPowerOffTimer:TimerStarted() then
		--Power Off in process, keep polling until no response
		--then OFF Notification will be sent to proxy in OnPowerOffTimerExpired()
		
	     gPowerOffTimer:StartTimer(2, "SECONDS")		
		gQueryPowerDelayTimer:StartTimer(1, "SECONDS")
		
		return
	end
	
	if (value == "ON") then
	     if gPowerOnTimer:TimerStarted() then
		  gQueryPowerDelayTimer:KillTimer()
		  gPowerOnTimer:KillTimer()
		  GetDeviceVolumeStatus()
		end
		
		if gSetInputDelayTimer:TimerStarted() then
		  gSetInputDelayTimer:StartTimer(2, "SECONDS")
		end
		
		if gSetChannelDelayTimer:TimerStarted() then
		  gSetChannelDelayTimer:StartTimer(4, "SECONDS")
		end		
		
		gProjectorProxy:dev_PowerOn()
		
	elseif (value== "OFF") then	--NOTE: TV does notrespond when off, notitication will be sent from OnPowerOffTimerExpired()
		gProjectorProxy:dev_PowerOff()
	else
		LogWarn("DEV_MSG.GET_POWER_STATUS(): value not valid - " .. value)
	end
end

function DEV_MSG.GET_CURRENT_APP(value)
    local pattern = ""
    
    --[[
    if (string.find(value, "com.webos.app") ~= nil) then
	   if (string.find(value, "externalinput") ~= nil) then
		  pattern = "APP:com.webos.app.externalinput.(.+)"
	   else
		  pattern = "APP:com.webos.app.(.+)"
		  --com.webos.app.livetv
	   end
    else
	   pattern = "APP:(.+)"
    end
    --]]
    
    pattern = "APP:(.+)"
    local app_name = string.match(value, pattern)
    
    local input_name = ""
    if (tInputResponseMap_IP[app_name] ~= nil) then 
	   input_name = tInputResponseMap_IP[app_name]
    else
	   if (gInputSelection.InProgress == true) then
		  PackAndQueueCommand("EXIT", "KEY_ACTION exit")
	   else
		  LogTrace("DEV_MSG.GET_CURRENT_APP(), " .. app_name .. " is currently selected.") 
		  return 
	   end
    end	
    
    LogTrace("DEV_MSG.GET_CURRENT_APP(), value = " .. value)
    
    local input
    if (Properties["Tuner Region"] == "Japan Model" and value == "02") then 
	   --['02']='BS (Broadcasting Satellite)',		--JAPAN
	   input = tInputConnMapByName['BS (Broadcasting Satellite)'].ID 
    else
	   input = tInputConnMapByName[input_name].ID 
    end

    local bindingID = tInputConnMapByID[input].BindingID
    LogInfo("INPUT_CHANGED, input = " .. input_name)

    --[[
    NOTE: Updated v2.10 (bug has existed for a long time but this is my current understanding...)
    There is TV proxy bug that affects Current Input in Composer Programming. 
    If connection has audio then it is considered audio, ONLY send audio connection notification.
    For video only connections (i.e. HDMI), BOTH audio and video connection notifications must be sent.
    --]]
    
    --all connections
    gProjectorProxy:dev_InputChanged(bindingID, input + 3000)		--audio
    
    --video only connections
    if (input > 20) then
	   gProjectorProxy:dev_InputChanged(bindingID, input + 1000)		--video
    end
    
end

function DEV_MSG.GET_VOLUME_STATUS(msg)

     local pattern = "(.+):(.+)"
     local _, value = string.match(msg, pattern)
	
	if (value == nil) then 
	   LogTrace("DEV_MSG.GET_VOLUME_STATUS(), exiting, value not valid: " .. msg)
	   return
	end
	
	LogTrace("DEV_MSG.GET_VOLUME_STATUS(), value = " .. value)
	
	local output = 0 --mod 1000 value of Output Connection ID
     deviceLevel = tonumber(value)
	local c4Level = deviceLevel
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	
	if (gMuted == "True") then
	   gProjectorProxy:dev_MuteChanged(0, "False")
	   gMuted = "False"
	end
end

function DEV_MSG.GET_MUTE_STATUS(value)
	LogTrace("DEV_MSG.GET_MUTE_STATUS(), value = " .. value)
	
	local output = 0  --mod 1000 value of Output Connection ID

	local state   	--Mute state represented as "True" or "False"
	if (value == "MUTE:on") then
		state = "True"
	elseif (value == "MUTE:off") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	
	if (gMuted ~= state) then
	   gProjectorProxy:dev_MuteChanged(output, state)
     end
	gMuted = state
	
end

function DEV_MSG.GET_CURRENT_CHANNEL(value)
    LogTrace("GET_CURRENT_CHANNEL(), value = " .. value)
    
    local pattern = ""
    if (string.find(value, " ")) then
	   pattern = "(.+) (.+)"
    else
	    pattern = "(.+)"
    end
    
    local channel_name, channel = string.match(value, pattern)
    print(channel_name, tostring(channel))
    
    if (channel_name == "NONE") then return end
    
    gProjectorProxy:dev_ChannelStatus(channel)
    --LogTrace("DEV_MSG.GET_CURRENT_CHANNEL(), channel = " .. channel)
end

--============================SERIAL CONTROL===================================--
function DEV_MSG.INPUT(value)
    if (tInputResponseMap[value] == nil) then return end	--nonsense in the response buffer most likely from when TV is in off state...
    
    LogTrace("DEV_MSG.INPUT(), value = " .. value)
    
    local input
    if (Properties["Tuner Region"] == "Japan Model" and value == "02") then 
	   --['02']='BS (Broadcasting Satellite)',		--JAPAN
	   input = tInputConnMapByName['BS (Broadcasting Satellite)'].ID 
    else
	   input = tInputConnMapByName[tInputResponseMap[value]].ID 
    end

    local bindingID = tInputConnMapByID[input].BindingID
    LogInfo("INPUT_CHANGED, input = " .. tInputResponseMap[value])

    --[[
    NOTE: Updated v2.10 (bug has existed for a long time but this is my current understanding...)
    There is TV proxy bug that affects Current Input in Composer Programming. 
    If connection has audio then it is considered audio, ONLY send audio connection notification.
    For video only connections (i.e. HDMI), BOTH audio and video connection notifications must be sent.
    --]]
    
    --all connections
    gProjectorProxy:dev_InputChanged(bindingID, input + 3000)		--audio
    
    --video only connections
    if (input > 20) then
	   gProjectorProxy:dev_InputChanged(bindingID, input + 1000)		--video
    end
    
end

function DEV_MSG.POWER(value)
	LogTrace("DEV_MSG.POWER(), value = " .. value)

	-- TODO: 01 & 00 values will need to be edited based upon the device protocol values
	--indicating if the device is on or off
	if (value == "01") then
	     gPowerOnTimer:KillTimer()
		
		if gSetInputDelayTimer:TimerStarted() then
		  gSetInputDelayTimer:StartTimer(2, "SECONDS")
		end
		
		if gSetChannelDelayTimer:TimerStarted() then
		  gSetChannelDelayTimer:StartTimer(2, "SECONDS")
		end	
		
		gProjectorProxy:dev_PowerOn()
	elseif (value== "00") then
		gProjectorProxy:dev_PowerOff()
	else
		LogWarn("DEV_MSG.POWER(): value not valid - " .. value)
	end
end

function DEV_MSG.VOLUME(value)
     if (value == nil or value == "") then return end	--nonsense in the response buffer most likely from when TV is in off state...
	
	LogTrace("DEV_MSG.VOLUME(), value = " .. value)
	
	local output = 0 --mod 1000 value of Output Connection ID
	local deviceLevel = tonumber(value, 16)
	local c4Level = deviceLevel
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	
	if (gMuted == "True") then
	   gProjectorProxy:dev_MuteChanged(0, "False")
	   gMuted = "False"
	end
end

function DEV_MSG.MUTE(value)
	LogTrace("DEV_MSG.MUTE(), value = " .. value)
	
	local output = 0  --mod 1000 value of Output Connection ID

	local state   	--Mute state represented as "True" or "False"
	if (value == "00") then
		state = "True"
	elseif (value == "01") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	
	if (gMuted ~= state) then
	   gProjectorProxy:dev_MuteChanged(output, state)
     end
	gMuted = state
	
end

function DEV_MSG.CHANNEL_EMEA(value) -- To do
    local high = tonumber(string.sub(value, 8, 9), 16)
    local low = tonumber (string.sub(value, 10, 11), 16)
    local channel_type = string.sub(value, 12, 13)
    
    -- Probably should do something here based on channel_type to see if we've changed RF input
    local channel = high * 256 + low

    gProjectorProxy:dev_ChannelStatus(channel)
    LogTrace("DEV_MSG.CHANNEL_EMEA(), channel = " .. channel)
end


function DEV_MSG.CHANNEL_US(value) -- To do
    local physical=string.sub(value, 8, 9)
    local major=string.sub(value, 12, 13)
    local minor=string.sub(value, 16, 17)
    local channel_type=string.sub(value, 18, 19)

    if (minor=='00') then
	   channel = tonumber (physical, 16)
    else
	   major=tonumber(major, 16)
	   minor=tonumber(minor, 16)
	   channel=major .. '-' .. minor
    end

    gProjectorProxy:dev_ChannelStatus(channel)
    LogTrace("DEV_MSG.CHANNEL_US(), channel = " .. channel)
end
 end)
package.preload['device_specific_commands'] = (function (...)
--[[=============================================================================
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_specific_commands = "2015.03.31"
end

--[[=============================================================================
    ExecuteCommand Code

    Define any functions for device specific commands (EX_CMD.<command>)
    received from ExecuteCommand that need to be handled by the driver.
===============================================================================]]

function EX_CMD.LaunchApp(tParams)
    if (gControlMethod ~= "Network") then print("Launch App only available with Network control!") return end
    LogTrace("EX_CMD.LaunchApp")
    LogTrace(tParams)
    
    local command = tParams["App"]
    PackAndQueueCommand("LaunchApp" , command)
end

--NOTE: If LG protocol supports a query to return all apps & APP_LAUNCH ids, then this table couold be built dynamcially
gAppLaunchCommandMap = {
    ['Amazon']='APP_LAUNCH amazon',
    ['Channel Plus']='APP_LAUNCH channelplus.us',
    ['Device Connector']='APP_LAUNCH com.webos.app.connectionwizard',
    ['Google Play']='APP_LAUNCH googleplaymovieswebos',
    ['Hulu']='APP_LAUNCH com.hulu.hulu',
    ['HuluPLUS']='APP_LAUNCH hulu-simplesmart',
    ['LG Content Store']='APP_LAUNCH com.webos.app.discovery',
    ['Multi-view']='APP_LAUNCH com.webos.app.multiview',
    ['Music']='APP_LAUNCH com.webos.app.music',
    ['Netflix']='APP_LAUNCH netflix',
    ['Photo & Video']='APP_LAUNCH com.webos.app.photovideo',
    ['Screen Share']='APP_LAUNCH com.webos.app.miracast',
    ['Sling TV']='APP_LAUNCH com.movenetworks.app.sling-tv-sling-production',
    ['Speaker Mode']='APP_LAUNCH com.webos.app.btspeakerapp',
    ['User Guide']='APP_LAUNCH com.webos.app.tvuserguide',
    ['viewster']='APP_LAUNCH com.lge.app.viewster',
    ['VUDU']='APP_LAUNCH vudu',
    ['Web Browser']='APP_LAUNCH com.webos.app.browser',
    ['YouTube']='APP_LAUNCH youtube.leanback.v4',
}

--called from 'Web Run App' Device Specific command to build app list dialog
function GetAppsForProgramming ()
    LogTrace('Getting Apps...')
    
    local list = getSortedList(gAppLaunchCommandMap)
    return list

end

function getSortedList(t)
	local a = {}
	for j in pairs(t) do 
		table.insert(a,j)
	end 
	table.sort(a)
    
    local list = {}

    for i = 1,table.maxn(a) do
        local id = a[i]
		table.insert (list, {text = id, value = t[id]})
    end
	
    return list
end

function EX_CMD.AspectRatio(tParams)
	LogTrace("EX_CMD.Aspect Ratio")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Aspect Ratio')
end

function EX_CMD.AVMode(tParams)
	LogTrace("EX_CMD.AVMode")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'AV Mode')
end

function EX_CMD.FavoriteChannel(tParams)
	LogTrace("EX_CMD.FavoriteChannel")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Favorite Channel')
end

function EX_CMD.Flashback(tParams)
	LogTrace("EX_CMD.Flashback")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Flashback')
end

function EX_CMD.ProgramGuide(tParams)
	LogTrace("EX_CMD.ProgramGuide")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Program Guide')
end

function EX_CMD.SIMPLINK(tParams)
	LogTrace("EX_CMD.SIMPLINK")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'SIMPLINK')
end

function EX_CMD.Soccer(tParams)
	LogTrace("EX_CMD.Soccer")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Soccer')
end

function EX_CMD.Toggle3D(tParams)
	LogTrace("EX_CMD.Toggle3D")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, '3D')
end

function EX_CMD.UserGuide(tParams)
	LogTrace("EX_CMD.UserGuide")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'User Guide')
end

function EX_CMD.Settings(tParams)
	LogTrace("EX_CMD.Settings")
	SEND_COMMAND_FROM_COMMAND_TABLE(PROJECTOR_PROXY_BINDINGID, 'Settings')
end

function EX_CMD.Volume(tParams)
     LogTrace("EX_CMD.SetVolume, Level = " .. tParams.Level)
     if (gControlMethod == "IR") then LogWarn("Set Volume Level is not supported via IR control!") return end
	
	SET_VOLUME_LEVEL(tParams.Level)
end

function EX_CMD.EnergySaving(tParams)
    --ENERGY_SAVING maximum | medium | minimum
    local mode = tParams["Mode"]
    LogTrace("EX_CMD.EnergySaving, Mode: " .. tostring(mode))
    
    PackAndQueueCommand("EnergySaving" , "ENERGY_SAVING " .. mode)
end

function EX_CMD.PictureMode(tParams)
    --PICTURE_MODE vivid | normal | game | cinema | sports
    local mode = tParams["Mode"]
    LogTrace("EX_CMD.PictureMode, Mode: " .. tostring(mode))
    
    PackAndQueueCommand("PictureMode" , "PICTURE_MODE " .. mode)
end

function EX_CMD.AudioBalance(tParams)
    --AUDIO_BALANCE [0 to 100]
    local level = tParams["Level"]
    LogTrace("EX_CMD.AudioBalance , Level: " .. level)
    
    PackAndQueueCommand("AudioBalance" , "AUDIO_BALANCE " .. level)
end
 end)
package.preload['actions'] = (function (...)
--[[=============================================================================
    Lua Action Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.actions = "2015.03.02"
end

-- TODO: Create a function for each action defined in the driver

function LUA_ACTION.TemplateVersion()
	TemplateVersion()
end

function LUA_ACTION.GetMACAddresses()
    GetMACAddressWired()
    GetMACAddressWifi()
end

function LUA_ACTION.ACTION_SetDeviceVisibility(tParams)
    LogTrace("Received ACTION Set Device Visibility")
    LogTrace(tParams)

    if (gProxyNameToBindingIdMap[tParams.Proxy] ~= nil) then
      local proxyBindingID = gProxyNameToBindingIdMap[tParams.Proxy]
      set_proxy_visbility(proxyBindingID, tParams.Action)
    else
      LogWarn("ACTION_SetDeviceVisibility(), the " .. tParams.Proxy .. " proxy is not included in this driver!")
    end
end end)
package.preload['common.c4_utils'] = (function (...)
--[[=============================================================================
    Helper functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_utils = "2016.01.08"
end

--[[=============================================================================
    AsciiToBCD(InString)

    Description
    Convert an ascii string to a binary coded decimal. Each decimal digit is
    stored in one byte, with the lower four bits encoding the digit in BCD form.

    Parameters
    InString(string) - Ascii string that is to be converted into bcd

    Returns
    The binary coded decimal
===============================================================================]]
function AsciiToBCD(InString)
	local WorkVal = 0
	local RetValStr = ""
	local DoingHighNybble = false
	local WorkStr = ((#InString % 2) == 0) and (InString) or ("0" .. InString)	-- make sure length is an even number

	for CharCount = 1, #WorkStr do
		local NumVal = tonumber(WorkStr:sub(CharCount, CharCount))

		WorkVal = bit.lshift(WorkVal, 4) + NumVal
		if (DoingHighNybble) then
			RetValStr = RetValStr .. string.char(WorkVal)
			WorkVal = 0
		end

		DoingHighNybble = (not DoingHighNybble)
	end

	return RetValStr
end

--[[=============================================================================
    BCDToAscii(InByte)

    Description
    Convert an BCD string to an ascii string.

    Parameters
    InByte(string) - Binary coded decimal that is to be converted into ascii

    Returns
    The ascii string
===============================================================================]]
function BCDToAscii(InByte)
	return tostring(bit.rshift(InByte, 4)) .. tostring(bit.band(InByte, 0x0F))
end

--[[=============================================================================
    MakeXMLNode(Tag, Value)

    Description
    Create an Xml element

    Parameters
    Tag(string)   - The Xml elements name
    Value(string) - The Xml elements value

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLNode(Tag, Value)
	return "<" .. Tag .. ">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)

    Description
    Create an Xml element with an attribute

    Parameters
    Tag(string)       - The Xml elements name
    Value(string)     - The Xml elements value
    Attribute(string) - The attribute to be added to the Xml element
    AttrValue(string) - The value of the attribute to be added

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)
    return "<" .. Tag .. " " .. Attribute .. "=\"" .. AttrValue .. "\">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    StringFromUnicode(UnicodeString)

    Description
    Convert a unicode string

    Parameters
    UnicodeString(string) - The unicode string to be converted to ascii

    Returns
    The ascii representation of the unicode string
===============================================================================]]
function StringFromUnicode(UnicodeString)
	local RetVal = ""

	-- extract every other byte from the unicode string
	for Index = 2, #UnicodeString, 2 do
		RetVal = RetVal .. string.sub(UnicodeString, Index, Index)
	end

	return RetVal
end

--[[=============================================================================
    StringSplit(s)

    Description
    Splits a string into multiple strings at an optionally specified delimiter
	If the delimiter is not specified, it will defalt to the space character

    Parameters
    s(string) - The string that is to be split into several strings
	d(string) - The delimiter to split the string on

    Returns
    A table of strings containing all the seperate values in the given string
===============================================================================]]
function StringSplit(s, d)
	local delim = (d ~= nil) and d or " "
	local result = {}

	if s == nil or s == "" then
		return result
	end

	for match in (s..delim):gmatch("(.-)"..delim) do
		table.insert(result, match)
	end

	return result
end

--[[=============================================================================
    toboolean(s)

    Description
    Returns a boolean representation of the given value

    Parameters
    val input value, may be of different types

    Returns
    The value true or false based on the given value
		If the value is of type string the return true if the first letter is "T" or "t" or if the string is "1"
		If the value is of type number the return true if the value is non-zero
		If the value is already a boolean, just return it.
===============================================================================]]
function toboolean(val)
	local rval = false;

	if type(val) == "string" and (string.lower(val) == 'true' or val == "1") then
		rval = true
	elseif type(val) == "number" and val ~= 0 then
		rval =  true
	elseif type(val) == "boolean" then
		rval = val
	end

	return rval
end

--[[=============================================================================
    tointeger(s)

    Description
    Force a number or a string representation of a number to be an integer

    Parameters
    val - A number or a string representation of a number

    Returns
    The the rounded off integer value.
===============================================================================]]
function tointeger(val)
	local nval = tonumber(val)
	return (nval >= 0) and math.floor(nval + 0.5) or math.ceil(nval - 0.5)
end


--[[=============================================================================
    Go(to, err, ...)

    Description
    Call a function with the given arguments if it exists or report the error

    Parameters
    to(string)  - The string to evaluate the boolean representation from
    err(string) - The error to report if the function does not exist
    ...         - Additional optional parameters for the function specified by
                  the "to" parameter

    Returns
    Nothing
===============================================================================]]
function Go(to, err, ...)
	if (type(to) == "function") then
		return to(...)
	else
		LogTrace(err)
	end
end

--[[=============================================================================
    IsEmpty(str)

    Description
    Identifies if the string given is nil or empty

    Parameters
    str(string) - The string to evaluate for the empty condition

    Returns
    True if the given value is empty, false otherwise
===============================================================================]]
function IsEmpty(str)
	return str == nil or str == ""
end

--[[=============================================================================
    ReverseTable(a)

    Description
    Reverse table entries (key=value, value=key)

    Parameters
    a(table) - The table to reverse

    Returns
    new reversed table
===============================================================================]]
function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function tonumber_loc(str, base)
  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

--[[=============================================================================
    HexToString(InString)

    Description
    Converts a string of Hex characters to a readable string of ASCII characters

    Parameters
    InString(string) - The string to be converted

    Returns
    A string showing the hex bytes of the InString
===============================================================================]]
function HexToString(InString)
	local RetVal = ""

	for Index = 1, #InString do
		RetVal = RetVal .. string.format("%02X ", InString:byte(Index))
	end
	return RetVal
end


--[[=============================================================================
    StringToHex(InString)

    Description
    Converts a string of ASCII characters to as string with the actual Hex bytes in them.
	Basically an array of hex bytes.

    Parameters
    InString(string) - The string to be converted

    Returns
    A string of hex bytes (really an array of hex values) 
===============================================================================]]
function StringToHex(InString)
	local RetVal = ""

	for HexByteString in string.gfind(InString, "%x%x") do
		RetVal = RetVal .. string.char(tonumber(HexByteString, 16))
	end
	return RetVal
end

function RecordHistory(severity, eventType, category, subcategory, description)
	C4:RecordHistory(severity, eventType, category, subcategory, description)
end

function RecordCriticalHistory(eventType, category, subcategory, description)
	RecordHistory("Critical", eventType, category, subcategory, description)
end

function RecordWarningHistory(eventType, category, subcategory, description)
	RecordHistory("Warning", eventType, category, subcategory, description)
end

function RecordInfoHistory(eventType, category, subcategory, description)
	RecordHistory("Info", eventType, category, subcategory, description)
end


 end)
package.preload['common.c4_ir_connection'] = (function (...)
--[[=============================================================================
    Base for an IR connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_ir_connection = "2016.01.08"
end

IRConnectionBase = inheritsFrom(DeviceConnectionBase)

function IRConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function IRConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("tSerConBase:Initialize")
	gControlMethod = "IR"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function IRConnectionBase:ControlMethod()
	return "IR"
end

function IRConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendIR(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("IR connection is not bound. Command not sent.")
	end
end
 end)
package.preload['common.c4_serial_connection'] = (function (...)
--[[=============================================================================
    Base for a serial connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_serial_connection = "2016.01.08"
end

SerialConnectionBase = inheritsFrom(DeviceConnectionBase)

function SerialConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function SerialConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "Serial"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function SerialConnectionBase:ControlMethod()
	return "Serial"
end

function SerialConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendToSerial(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("Not connected to serial. Command not sent.")
	end
end

function SerialConnectionBase:SendRaw(sData)
	C4:SendToSerial(self._BindingID, sData)
end


function SerialConnectionBase:ReceivedFromSerial(idBinding, sData)
	self:ReceivedFromCom(sData)
end
 end)
package.preload['lib.c4_queue'] = (function (...)
--[[=============================================================================
    c4_queue Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_queue = "2016.01.08"
end

c4_queue = inheritsFrom(nil)

function c4_queue:construct()
	-- entry table
	self._et = {first = 0, last = -1}
	self._maxSize = 0	-- no size limit
	self._name = ""

	local mt = getmetatable(self)
	if (mt ~= nil) then
		mt.__tostring = self.__tostring
	end
end

function c4_queue:__tostring()
	local tOutputString = {}
	table.insert(tOutputString, "--- Queue ---")
	if (not IsEmpty(self._name)) then
		table.insert(tOutputString, "  name = " .. tostring(self._name))
	end
	table.insert(tOutputString, "  first = " .. tostring(self._et.first))
	table.insert(tOutputString, "  last = " .. tostring(self._et.last))
	table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
	table.insert(tOutputString, "  maximum size = " .. self._maxSize)
	table.insert(tOutputString, "  next value = " .. tostring(self:value()))
	return table.concat(tOutputString, "\n")
end


-- push a value on the queue
function c4_queue:push(value, ...)
	local numItems = self._et.last - self._et.first + 1

	if ((self._maxSize <= 0) or (numItems < self._maxSize) ) then
		local last = self._et.last + 1
		self._et.last = last
		
		local interval = select(1, ...)
		local units = select(2, ...)
		local command_name = select(3, ...)
		self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}		
		--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
	else
		-- if addToQueue == true then push value to queue
		if (self:OnMaxSizeReached()) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}
			--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		end
	end
end

function c4_queue:OnMaxSizeReached()
	--LogTrace ("Max Size Reached - clear queue and push value to the queue (default).")
	local addToQueue = true

	self:clear()
	return (addToQueue)
end

-- pop a value from the queue
function c4_queue:pop()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace("Queue:pop(), queue is empty")
		return ""
	end

	local value = self._et[first]
	self._et[first] = nil        -- to allow garbage collection
	self._et.first = first + 1
	--LogTrace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)

	return value
end

-- clear queue
function c4_queue:clear()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace ("Queue:clear(), queue is empty")
		return ""
	end

	self._et = {first = 0, last = -1}
	--LogTrace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
	--LogTrace (self._et)
end

-- return value of first item
function c4_queue:value()
	local first = self._et.first

	if (first > self._et.last) then
		return ""
	else
		return self._et[first]
	end
end

-- return queue's maximum size
function c4_queue:MaxSize()
	return self._maxSize
end

-- return queue's maximum size
function c4_queue:SetMaxSize(size)
	self._maxSize = size
end

function c4_queue:SetName(name)
	self._name = name
end

-- return the queue's current size
function c4_queue:Size()
	return self._et.last - self._et.first + 1
end

-- is queue empty?
function c4_queue:empty()
	-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
	if (self._et.first > self._et.last) then
		return true
	else
		return false
	end
end

--[[
    c4_queue unit tests
--]]
function __test_c4_queue()
	require "test.C4Virtual"
	require "lib.c4_log"

	local LOG = c4_log:new("test_c4_queue")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	-- create an instance of the queue
	local c4Queue = c4_queue:new()

	c4Queue:SetMaxSize(3)
	assert(c4Queue:MaxSize() == 3, "_maxSize is not equal to '3' it is: " .. c4Queue:MaxSize())

	c4Queue:push("Item #1 in Queue")
	c4Queue:push("Item #2 in Queue")
	c4Queue:push("Item #3 in Queue")
	c4Queue:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4Queue:Size() == 1, "queue size is not equal to '1' it is: " .. c4Queue:Size())

	print (c4Queue)

	-- Test inheritance overriding OnMaxSizeReached
	-- Create a new class
	c4_queue_new = inheritsFrom(c4_queue)

	-- override construct()
	function c4_queue_new:construct()
		self.superClass():construct()  -- call base class
		self._maxSizeOption = 1

		local mt = getmetatable(self)
		if (mt ~= nil) then
			mt.__tostring = self.__tostring
		end
	end

	-- override OnMaxSizeReached()
	function c4_queue_new:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,

		-- Option 1: Do Nothing, new item is not added to queue
		if (self._maxSizeOption == 1) then
			
			LogInfo("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)
		-- Option 2: pop value, and push new value on queue
		elseif(self._maxSizeOption == 2) then
			LogInfo("Max Size Reached - pop value, and push new value on queue (option 2)")
			self:pop()
			return (true)
		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(self._maxSizeOption == 3) then
			LogInfo("Max Size Reached - clear queue and DO NOT push new value onto queue")
			self:clear()
			return (false)
		end
	end

	-- create an instance of the new queue
	local c4QueueNew = c4_queue_new:new()
	c4QueueNew:SetMaxSize(3)
	c4QueueNew:push("Item #1 in Queue")
	c4QueueNew:push("Item #2 in Queue")
	c4QueueNew:push("Item #3 in Queue")
	c4QueueNew:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4QueueNew:Size() == 3, "queue size is not equal to '3' it is: " .. c4QueueNew:Size())

	print(c4QueueNew)
	print ("done...")
end end)
package.preload['common.c4_device_connection_base'] = (function (...)
--[[=============================================================================
    DeviceConnectionBase Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_common"
require "lib.c4_object"
require "lib.c4_log"
require "lib.c4_timer"
require "lib.c4_queue"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_device_connection_base = "2016.01.08"
end

COMMAND_QUEUE_SIZE = 100
DEFAULT_COMMAND_DELAY_INTERVAL = 100            -- Don't send consecutive commands faster than this many milliseconds
DEFAULT_COMMAND_RESPONSE_INTERVAL = 3           -- If we haven't received and ACK after this many seconds, try again
DEFAULT_RETRY_COUNT_MAX = 3

function ON_DRIVER_EARLY_INIT.c4_device_connection_base()
	gReceiveBuffer = ""
	gIsUrlConnected = false
	gIsNetworkServerConnected = false
	gIsNetworkConnected = false
	gIsSerialConnected = false
	gIsIRConnected = false
end

DeviceConnectionBase = inheritsFrom(nil)

function DeviceConnectionBase:construct()

	self._IsConnected = false
	self._SendTimer = nil
	self._WaitResponseTimer = nil
	self._CommandQueue = nil
	self._Priority1CommandQueue = nil
	self._Priority2CommandQueue = nil
	self._LastCommand = nil
	self._ExpectAck = false
	self._CommandRetryCount = 0
	self._RetryCountMax = DEFAULT_RETRY_COUNT_MAX

	self._SendCommandDelayMS = DEFAULT_COMMAND_DELAY_INTERVAL
	self._CommandResponseWaitS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	-- Polling
	self._PollingInterval = 0
	self._PollingUnits = "SECONDS"
	self._PollingTimer = nil
end

function DeviceConnectionBase:Initialize(ExpectAck, CommandDelayInterval, CommandResponseInterval, CallbackParam)

	if (ExpectAck ~= nil) then
		self._ExpectAck = ExpectAck
	end

	if (CommandDelayInterval ~= nil) then
		self._SendCommandDelayMS = CommandDelayInterval
	end

	if (CommandResponseInterval ~= nil) then
		self._CommandResponseWaitS = CommandResponseInterval
	end

	self._CommandQueue = c4_queue:new()
	self._CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._CommandQueue:SetName("Command Queue")

	self._Priority1CommandQueue = c4_queue:new()
	self._Priority1CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority1CommandQueue:SetName("P1 Queue")

	self._Priority2CommandQueue = c4_queue:new()
	self._Priority2CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority2CommandQueue:SetName("P2 Queue")

	-- usually only one of these timers will be used, but it's pretty low overhead to instantiate both of them
	self._SendTimer = c4_timer:new("SendCommand", self._SendCommandDelayMS, "MILLISECONDS", DeviceConnectionBase.OnSendTimeExpired, false, CallbackParam)
	self._WaitResponseTimer = c4_timer:new("WaitResponse", self._CommandResponseWaitS, "SECONDS", DeviceConnectionBase.OnWaitTimeExpired, false, CallbackParam)
end

function DeviceConnectionBase:InitPolling(PollingInterval, PollingUnits, CallbackParam)
	LogFatal("DeviceConnectionBase:InitPolling()")
	if (PollingInterval ~= nil) then
		self._PollingInterval = PollingInterval
	end
	
	self._PollingUnits = PollingUnits or self._PollingUnits

	LogFatal("self._PollingInterval: %s, self._PollingUnits: %s", tostring(self._PollingInterval), tostring(self._PollingUnits))
	
	-- create polling timer
	self._PollingTimer = c4_timer:new("Polling", self._PollingInterval, self._PollingUnits, DeviceConnectionBase.OnPollingTimerExpired, false, CallbackParam)
end

function DeviceConnectionBase:StartPolling(interval, units)
	LogFatal("DeviceConnectionBase:StartPolling()")
	LogFatal("self._PollingTimer: %s", tostring(self._PollingTimer))
	
	if (self._PollingTimer ~= nil) then
		self._PollingTimer:KillTimer()
		
		local timer_units = units or self._PollingTimer._units
		local timer_interval = interval or self._PollingInterval

		self._PollingTimer:StartTimer(timer_interval, timer_units)
	end
end

function DeviceConnectionBase:StopPolling()
	LogFatal("DeviceConnectionBase:StopPolling()")
	self._PollingTimer:KillTimer()
end

function DeviceConnectionBase:SetExpectACK(ExpectACK)
	self._ExpectAck = ExpectACK
end

function DeviceConnectionBase:SetCommandDelayInterval(DelayInterval)
	self._SendCommandDelayMS = DelayInterval
end

function DeviceConnectionBase:SetResponseWaitInterval(WaitInterval)
	self._CommandResponseWaitS = WaitInterval
end

function DeviceConnectionBase:ReceivedFromCom(sData)

	gReceiveBuffer = gReceiveBuffer .. sData
	LogTrace("ReceivedFromCom  ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)

	message = self:GetMessage()
	while (message ~= nil and message ~= "") do
		status, err = pcall(HandleMessage, message)
		if (status) then
			message = self:GetMessage()
		else
			LogError("LUA_ERROR: " .. err)
			message = ""
			gReceiveBuffer = ""
		end
	end
end

function DeviceConnectionBase:SetConnection(IsConnected, method)
	self._IsConnected = IsConnected
	gControlMethod = method
end

function DeviceConnectionBase:ControlMethod()
	-- Override in derived class
	print("WARNING: Need to override ControlMethod - should never be called")
	
	return ""
end

function DeviceConnectionBase:StartCommandTimer(...)
	local value = select(1, ...)
	local units = select(2, ...)
	local command_name = select(3, ...) or ""

	self._WaitResponseTimer:KillTimer()
	self._SendTimer:KillTimer()

	if (self._ExpectAck) then
		-- expecting an ACK set the Response Wait timer
		local timer_units = units or self._WaitResponseTimer._units
		local timer_interval = value or self._CommandResponseWaitS

		self._WaitResponseTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting wait Timer:  %d", self._WaitResponseTimer._timerID) .. " for " .. command_name)
	else
		-- no ACK expected, just wait the designated amount of time and send another command
		local timer_units = units or self._SendTimer._units
		local timer_interval = value or self._SendCommandDelayMS

		self._SendTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting Send Timer:  %d for %s (timer_interval = %d, timer_units = %s)", self._SendTimer._timerID, command_name, timer_interval, timer_units))
	end
end

-- Note the '.' instead of the ':'
function DeviceConnectionBase.OnSendTimeExpired(Instance)
	LogTrace("Send Timer expired")
	Instance._SendTimer:KillTimer()

	local tCommand = Instance._LastCommand
	if (tCommand ~= nil) then
		if (tCommand.command_name ~= nil) then
			LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand.command_name))
		elseif (type(tCommand) == "string") then
			LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand))
		end
	else
		LogTrace("Send Timer expired - Last Command: UNKNOWN, Send Next Command")
	end
	
	Instance._LastCommand = nil
	Instance:SendNextCommand()

	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase.OnWaitTimeExpired(Instance)
    LogTrace("Wait Timer expired")
    Instance._WaitResponseTimer:KillTimer()
    Instance._CommandRetryCount = Instance._CommandRetryCount + 1
	
    local last_command_name = Instance._LastCommand.command_name

    if (Instance._CommandRetryCount >= Instance._RetryCountMax) then
	   -- To many retries, pop the current command and try the next one
	   Instance._CommandRetryCount = 0
	   Instance._LastCommand = nil
	   Instance:SendNextCommand()
    else
	   if (last_command_name == "ENTER") then
		  if gCheckPowerStateFromOnCommand then
			 gCheckPowerStateFromOnCommand = false
			 Instance._CommandRetryCount = 0
			 Instance._LastCommand = nil
			 Instance:SendNextCommand()
		  end
	   else
		  Instance:SendLastCommand()
	   end
	end
end

function DeviceConnectionBase.OnPollingTimerExpired(Instance)
	LogTrace("Polling Timer expired")
	Instance._PollingTimer:KillTimer()

	OnPollingTimerExpired()
	
	Instance._PollingTimer:StartTimer(Instance._PollingInterval)
end

function DeviceConnectionBase:HandleACK()
	self._LastCommand = nil
	
	self._WaitResponseTimer:KillTimer()
	self._CommandRetryCount = 0
	self:SendNextCommand()
	
	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase:QueueEmpty()
	return (self._CommandQueue:empty() and self._Priority1CommandQueue:empty() and self._Priority2CommandQueue:empty())
end

function DeviceConnectionBase:QueueCommand(sCommand, ...)
--	LogTrace("QueueCommand(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority1Command(sCommand, ...)
	LogTrace("QueuePriority1Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority1CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority2Command(sCommand, ...)
	LogTrace("QueuePriority2Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority2CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:SendNextCommand()
	LogTrace("DeviceConnectionBase:SendNextCommand")

	local tCommand = nil
	if (not self._Priority1CommandQueue:empty()) then
		tCommand = self._Priority1CommandQueue:pop()
		LogTrace(tostring(gCon._Priority1CommandQueue))
	elseif (not self._Priority2CommandQueue:empty()) then
		tCommand = self._Priority2CommandQueue:pop()
		LogTrace(tostring(gCon._Priority2CommandQueue))
	elseif (not self._CommandQueue:empty()) then
		tCommand = self._CommandQueue:pop()
		LogTrace(tostring(gCon._CommandQueue))
	end
	
	if (tCommand ~= nil) then
		self._LastCommand = tCommand
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendLastCommand()
--	LogTrace("DeviceConnectionBase:SendLastCommand")

	local tCommand = self._LastCommand
	if (tCommand ~= nil) then
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendCommand()
	-- Dummy routine.  Override in derived class
	print("Need to override SendCommand - should never be called")
end


function DeviceConnectionBase:GetMessage()
	
	-- Brain dead version of this routine. Just return the current receive buffer.
	-- It's very likely that a GetMessage() function will need to be created
	if (GetMessage ~= nil and type(GetMessage) == "function") then
		return GetMessage()
	else
		local ComMessage = gReceiveBuffer
		gReceiveBuffer = ""

		return ComMessage
	end
end

--[[=============================================================================
    Other Connection Functions
===============================================================================]]

function ReceivedFromSerial(idBinding, sData)
	if (gCon.ReceivedFromSerial == nil) then return end --serial is bound but not the current control method
	gCon:ReceivedFromSerial(idBinding, sData)
end

function ReceivedFromNetwork(idBinding, nPort, sData)
	gCon:ReceivedFromNetwork(idBinding, nPort, sData)
end

function OnServerDataIn(nHandle, strData)
--	LogTrace("Received Data on Handle: " .. nHandle .. ": " .. strData)
--	LogTrace("Data Is: %s", HexToString(strData))
	gCon:ReceivedFromNetworkServer(nHandle, strData)
end


--[[=============================================================================
    The ReceivedAsync function is called in response to 'url_get_request'. 
    The ticketId is the number returned from the request.
===============================================================================]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

--	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
--	LogTrace(tHeaders)

	gCon:ReceivedAsync(ticketId, strData, responseCode, tHeaders)
end
	
--[[=============================================================================
    OnBindingChanged(idBinding, class, bIsBound)
  
    Description:
    Function called by Director when a binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnBindingChanged(idBinding, class, bIsBound)
	
	LogTrace("OnBindingChanged(): idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	if (idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
		SetControlMethod()
		OnSerialConnectionChanged(idBinding, class, bIsBound)
	elseif (idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
		SetControlMethod()
		OnIRConnectionChanged(idBinding, class, bIsBound)
	elseif(OnConnectionChanged ~= nil and type(OnConnectionChanged) == "function") then
		OnConnectionChanged(idBinding, class, bIsBound)
	end
end

--[[=============================================================================
    OnNetworkBindingChanged(idBinding, bIsBound)
  
    Description:
    Function called by Director when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	LogTrace('OnNetworkBindingChanged(): idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))

	gIsNetworkConnected = bIsBound
	SetControlMethod()
	OnNetworkConnectionChanged(idBinding, bIsBound)
	if (bIsBound) then
		-- Start a special instance of reconnect timer to eventually do NetConnect if not done automatically
		--gCon._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval) 	
	end	
end

--[[=============================================================================
    OnConnectionStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. tostring(nPort) .. ")]: " .. sStatus)

	local isOnline = false

	gNetworkStatus = sStatus	
	if (sStatus == "ONLINE") then
		isOnline = true
	end

	gCon:SetOnlineStatus(isOnline)
	OnNetworkStatusChanged(idBinding, nPort, sStatus)
end

--[[=============================================================================
    SetControlMethod()
  
    Description:
    Sets the control method type for the drivers internal infrastructure
  
    Parameters:
    None
  
    Returns:
    The type of control method for the drivers connection(i.e. Network, Serial,
    IR, or (none))
===============================================================================]]
function SetControlMethod()
	if (gCon ~= nil) then
		if (gIsNetworkConnected == false) and (gCon._NetworkReconnectTimer ~= nil) then
			--housekeeping when changing from network control to serial or IR control
			gCon._NetworkReconnectTimer:KillTimer() 
		end
	end


	if( gIsNetworkServerConnected) then
		-- connect to NetworkServer communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "NetworkServer") then
			gCon = NetworkServerConnectionBase:new()
			gCon:Initialize()
		end
		gCon:SetConnection(true, "NetworkServer")
	elseif (gIsNetworkConnected) then
		-- connect to Network communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Network") then
			gCon = NetworkConnectionBase:new(NETWORK_BINDING_ID, NETWORK_PORT)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "Network")
	elseif (gIsUrlConnected) then
		-- connect to URL communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "URL") then
			gCon = UrlConnectionBase:new()
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "URL")
	elseif (gIsSerialConnected) then
		-- connect to Serial communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Serial") then
			gCon = SerialConnectionBase:new(SERIAL_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
			gCon:InitPolling(tonumber(gPollingTimerInterval), "MINUTES", gCon)
		end
		gCon:SetConnection(true, "Serial")
	elseif (gIsIRConnected) then
		-- connect to IR communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "IR") then
			gCon = IRConnectionBase:new(IR_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "IR")
	else
		if (gCon ~= nil) then
			gCon:SetConnection(false, "(none)")
		end
		-- gCon = nil
	end

	gCon._CommandQueue:clear()
	gCon._Priority1CommandQueue:clear()
	gCon._Priority2CommandQueue:clear()
end

--[[=============================================================================
    ValidateControlMethod(controlMethod)
  
    Description:
    Identifies whether the specified control method has a valid connection
  
    Parameters:
    controlMethod(string) - The communication we are validating against
                            Valid types are (Network, Serial, and IR)
  
    Returns:
    true if the controlMethod specified has been connected, false otherwise.
===============================================================================]]
function ValidateControlMethod(controlMethod)
	local isValid = false

	if (controlMethod == "Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (controlMethod == "URL") and (gIsUrlConnected) then
		isValid = true
	elseif (controlMethod == "Serial") and (gIsSerialConnected) then
		isValid = true
	elseif (controlMethod == "IR") and (gIsIRConnected) then
		isValid = true
	end

	return isValid
end
 end)
package.preload['common.c4_network_connection'] = (function (...)
--[[=============================================================================
    Base for a network connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"
require "lib.c4_log"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_network_connection = "2016.01.08"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30
DEFAULT_RECONNECT_INTERVAL_SECONDS = 5

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS
gNetworkReconnectInterval = DEFAULT_RECONNECT_INTERVAL_SECONDS

NetworkConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkConnectionBase:construct(BindingID, Port)
	self.superClass():construct()

	self._BindingID = BindingID
	self._Port = Port
	self._LastCheckin = 0
	self._IsOnline = false
	self._KeepAliveTimer = nil
end

function NetworkConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetConBase:Initialize")
	gControlMethod = "Network"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	self._KeepAliveTimer = c4_timer:new("PollingTimer", gNetworkKeepAliveInterval, "SECONDS", NetworkConnectionBase.OnKeepAliveTimerExpired, false, self)
	self._NetworkReconnectTimer = c4_timer:new("NetworkReconnectTimer", gNetworkReconnectInterval, "SECONDS", NetworkConnectionBase.OnNetworkReconnectTimerExpired, false, self)
end

function NetworkConnectionBase:ControlMethod()
	return "Network"
end

function NetworkConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)
			gDevMsgCallback = command_name
			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end

function NetworkConnectionBase:ReceivedFromNetwork(idBinding, nPort, sData)
	self._LastCheckin = 0
	self:ReceivedFromCom(sData)
end

function NetworkConnectionBase:CheckNetworkConnectionStatus()
	if (self._IsConnected and (not self._IsOnline)) then
		LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(self._BindingID, self._Port)
		--C4:NetConnect(self._BindingID, self._Port)
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	end
end

function NetworkConnectionBase.OnKeepAliveTimerExpired(Instance)
	Instance._LastCheckin = Instance._LastCheckin + 1

	if(Instance._LastCheckin == 3) then
		LogWarn("Failed to receive poll responses... initiating network recovery mode...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		return
	elseif(Instance._LastCheckin > 4) then	
		Instance._LastCheckin = 4
	end

	if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function" and Instance._IsOnline) then
		SendKeepAlivePollingCommand()
	end

	Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
end

function NetworkConnectionBase.OnNetworkReconnectTimerExpired(Instance)
     --PC 1/24/2018 added Instance._IsOnline test to stop reconnect when TV is OFF
	if Instance._IsConnected then
		LogWarn("OnNetworkReconnectTimerExpired: Attempting to reactivate network connection...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		C4:NetConnect(Instance._BindingID, Instance._Port)
		
		if (gProjectorProxy._PowerState ~= "OFF" or gPowerOnTimer:TimerStarted()) then
		  Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	     else
		  LogWarn("The TV is Powered Off, not restarting NetworkReconnectTimer.")
	     end
	else
		LogWarn("Cannot attempt to reactivate, the network connection is not bound.")
		LogWarn("gProjectorProxy._PowerState = ".. gProjectorProxy._PowerState)
	end
end

function NetworkConnectionBase:SetOnlineStatus(IsOnline)
	self._IsOnline = IsOnline

	if(IsOnline) then
		self._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
		self._NetworkReconnectTimer:KillTimer()
		self._LastCheckin = 0
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "true")
		end

		  --if WaitTimer is not running then a command may be stuck in the queue, send it again
		  if (self._WaitResponseTimer:TimerStarted() == false and self._SendTimer:TimerStarted() == false) then 
			 self:SendLastCommand() 
		  else
			 self:SendNextCommand()
		  end
	else
		self._KeepAliveTimer:KillTimer()
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "false")
		end
	end
end

function ON_DRIVER_LATEINIT.c4_network_connection()
	-- Ensure existing connection is taken into consideration (useful on Driver Update)
	if (gCon ~= nil) then
		local tmp = C4:GetBindingAddress(gCon._BindingID)
		if (tmp ~= nil and string.len(tmp) > 0) then OnNetworkBindingChanged(gCon._BindingID, true) end
	end
end
 end)
package.preload['common.c4_notify'] = (function (...)
--[[=============================================================================
    Notification Functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_notify = "2016.01.08"
end

--[[=============================================================================
    SendNotify(notifyText, tParams, bindingID)

    Description
    Forwards a notification to the proxy with a list of parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    tParams(table)     - Table of key value pairs that hold the the parameters
                         and their values used in the proxy function
    bindingID(int)     - The requests binding id

    Returns
    Nothing
===============================================================================]]
function SendNotify(notifyText, tParams, bindingID)
	C4:SendToProxy(bindingID, notifyText, tParams, "NOTIFY")
end

--[[=============================================================================
    SendSimpleNotify(notifyText, ...)

    Description
    Forwards a notification to the proxy with no parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    bindingID(int)     - Optional parameter containing the requests binding id,
                         if not specified then the DEFAULT_PROXY_ID is given.

    Returns
    Nothing
===============================================================================]]
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end end)
package.preload['common.c4_command'] = (function (...)
--[[=============================================================================
    Functions for handling and executing commands and actions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_command = "2016.01.08"
end

--[[=============================================================================
    ExecuteCommand(sCommand, tParams)

    Description
    Function called by Director when a command is received for this DriverWorks
    driver. This includes commands created in Composer programming.

    Parameters
    sCommand(string) - Command to be sent
    tParams(table)   - Lua table of parameters for the sent command

    Returns
    Nothing
===============================================================================]]
function ExecuteCommand(sCommand, tParams)
	LogTrace("ExecuteCommand(" .. sCommand .. ")")
	LogInfo(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")
	local status, ret

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		status, ret = pcall(EX_CMD[sCommand], tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		status, ret = pcall(EX_CMD[trimmedCommand], tParams)
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
		status = true
	else
		LogInfo("ExecuteCommand: Unhandled command = " .. sCommand)
		status = true
	end
	
	if (not status) then
		LogError("LUA_ERROR: " .. ret)
	end
	
	return ret -- Return whatever the function returns because it might be xml, a return code, and so on
end

--[[=============================================================================
    EX_CMD.LUA_ACTION(tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    tParams(table) - Lua table of parameters for the command option

    Returns
    Nothing
===============================================================================]]
function EX_CMD.LUA_ACTION(tParams)
	if (tParams ~= nil) then
		for cmd, cmdv in pairs(tParams) do
			if (cmd == "ACTION" and cmdv ~= nil) then
				local status, err = pcall(LUA_ACTION[cmdv], tParams)
				if (not status) then
					LogError("LUA_ERROR: " .. err)
				end
				break
			end
		end
	end
end

--[[=============================================================================
    ReceivedFromProxy(idBinding, sCommand, tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    idBinding(int)   - Binding ID of the proxy that sent a BindMessage to the
                       DriverWorks driver.
    sCommand(string) - Command that was sent
    tParams(table)   - Lua table of received command parameters

    Returns
    Nothing
===============================================================================]]
function ReceivedFromProxy(idBinding, sCommand, tParams)

	if (sCommand ~= nil) then

		-- initial table variable if nil
		if (tParams == nil) then
			tParams = {}
		end
		
		LogTrace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function PRX_CMD." .. sCommand .. "()")
		LogInfo(tParams)
		
	    --support for universal miniapps
	     if (idBinding == SWITCHER_PROXY and sCommand == 'PASSTHROUGH') then
		     idBinding = PASSTHROUGH_PROXY
		     sCommand = tParams.PASSTHROUGH_COMMAND
	     end

		if ((PRX_CMD[sCommand]) ~= nil) then
			local status, err = pcall(PRX_CMD[sCommand], idBinding, tParams)
			if (not status) then
				LogError("LUA_ERROR: " .. err)
			end
		else
			LogInfo("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

--[[
	This function is called when a UI (Navigator) requests data, and
	calls the function requested.
--]]
function UIRequest(sRequest, tParams)
	local ret = ""

	if (sRequest ~= nil) then
		tParams = tParams or {}   -- initial table variable if nil
		LogTrace("UIRequest(): " .. sRequest .. "; Call Function UI_REQ." .. sRequest .. "()")
		LogInfo(tParams)

		if (UI_REQ[sRequest]) ~= nil then
			ret = UI_REQ[sRequest](tParams)
		else
			LogWarn("UIRequest: Unhandled request = " .. sRequest)
		end
	end

	return ret
end
 end)
package.preload['common.c4_property'] = (function (...)
--[[=============================================================================
    Function for changing properties

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_property = "2016.01.08"
end

--[[=============================================================================
    OnPropertyChanged(sProperty)

    Description
    Function called by Director when a property changes value. The value of the
    property that has changed can be found with: Properties[sName]. Note that
    OnPropertyChanged is not called when the Property has been changed by the
    driver calling the UpdateProperty command, only when the Property is changed
    by the user from the Properties Page. This function is called by Director
    when a property changes value.

    Parameters
    sProperty(string) - Name of property that has changed.

    Returns
    Nothing
===============================================================================]]
function OnPropertyChanged(sProperty)
	local propertyValue = Properties[sProperty]

	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	end

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	local status = true
	local err = ""

	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[sProperty], propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[trimmedProperty], propertyValue)
	end

	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)
	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end
 end)
package.preload['common.c4_init'] = (function (...)
--[[=============================================================================
    Initial driver initialization and destruction functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_property"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_init = "2016.01.08"
end

--[[=============================================================================
    OnDriverInit()

    Description
    Invoked by director when a driver is loaded. This API is provided for the
    driver developer to contain all of the driver objects that will require
    initialization.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverInit()
	gInitializingDriver = true
	C4:ErrorLog("INIT_CODE: OnDriverInit()")

	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_EARLY_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property
	-- global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		C4:ErrorLog("INIT_CODE: Calling OnPropertyChanged - " .. k .. ": " .. v)
		local status, err = pcall(OnPropertyChanged, k)
		if (not status) then
			C4:ErrorLog("LUA_ERROR: " .. err)
		end
	end

	gInitializingDriver = false
end

--[[=============================================================================
    OnDriverLateInit()

    Description
    Invoked by director after all drivers in the project have been loaded. This
    API is provided for the driver developer to contain all of the driver
    objects that will require initialization after all drivers in the project
    have been loaded.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[=============================================================================
    OnDriverDestroyed()
    Function called by Director when a driver is removed. Release things this
    driver has allocated such as timers.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	
	-- Call all ON_DRIVER_DESTROYED functions
	for k, v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end end)
package.preload['lib.c4_timer'] = (function (...)
--[[=============================================================================
    c4_timer Class

    Copyright 2017 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_timer = "2017.05.03"
end

c4_timer = inheritsFrom(nil)

function c4_timer:construct(name, interval, units, Callback, repeating, CallbackParam)
	self._name = name
	self._timerID = TimerLibGetNextTimerID()
	self._interval = interval
	self._units = units
	self._repeating = repeating or false
	self._Callback = Callback
	self._CallbackParam = CallbackParam or ""
	self._id = 0

	gTimerLibTimers[self._timerID] = self
	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("Created timer " .. self._name)
	end
end

function c4_timer:StartTimer(...)
	c4_timer:KillTimer()

	-- optional parameters (interval, units, repeating)
	if ... then
		local interval = select(1, ...)
		local units = select(2, ...)
		local repeating = select(3, ...)

		self._interval = interval or self._interval
		self._units = units or self._units
		self._repeating = repeating or self._repeating
	end

	if (tonumber(self._interval) > 0) then
		if (LOG ~= nil and type(LOG) == "table") then
			LogTrace("Starting Timer: " .. self._name)
		end

		self._id = C4:AddTimer(self._interval, self._units, self._repeating)
	end
end

function c4_timer:KillTimer()
	if (self._id) then
		self._id = C4:KillTimer(self._id)
	end
end

function c4_timer:TimerStarted()
	return (self._id ~= 0)
end

function c4_timer:TimerStopped()
	return (self._id == 0)
end

function c4_timer:GetTimerInterval()
	return (self._interval)
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.c4_timer()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.c4_timer()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[=============================================================================
    OnTimerExpired(idTimer)

    Description:
    Function called by Director when the specified Control4 timer expires.

    Parameters:
    idTimer(string) - Timer ID of expired timer.
===============================================================================]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._CallbackParam)
			end
		end
	end
end

--[[=============================================================================
    CreateTimer(name, interval, units, callback, repeating, callbackParam)

    Description: 
    Creates a named timer with the given attributes

    Parameters:
    name(string)       - The name of the timer being created
    interval(int)      - The amount of the given time between calls to the
                         timers callback function
    units(string)      - The time of time interval used (e.g. MILLSECONDS, SECONDS, MINUTES, HOURS)
    callback(string)   - The function to call when the timer expires
    repeating(bool)    - Parameter indicating whether the timer should be
                         called repeatedly until cancelled
    callbackParam(...) - Parameters to be passed to the callback function

    Returns:
    A handle to the timer
===============================================================================]]
function CreateTimer(name, interval, units, callback, repeating, callbackParam)
	timer = c4_timer:new(name, interval, units, callback, repeating, callbackParam)
	return timer
end

--[[=============================================================================
    StartTimer(handle, ...)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer)   - Handle to a created timer object
    interval(int)   - The amount of the given time between calls to the
                      timers callback function
    units(string)   - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    repeating(bool) - Parameter indicating whether the timer should be
                      called repeatedly until cancelled

    Returns:
    None
===============================================================================]]
function StartTimer(handle, ...)
	handle:StartTimer(...)
end

--[[=============================================================================
    KillTimer(handle)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    None
===============================================================================]]
function KillTimer(handle)
	handle:KillTimer()
end

--[[=============================================================================
    TimerStarted(handle)

    Description: 
    Identifies whether a timer has been started or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been started, or false otherwise
===============================================================================]]
function TimerStarted(handle)
	return handle:TimerStarted()
end

--[[=============================================================================
    TimerStopped(handle)

    Description: 
    Identifies whether a timer has been stopped or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been stopped, or false otherwise
===============================================================================]]
function TimerStopped(handle)
	return handle:TimerStopped()
end

--[[=============================================================================
    GetTimerInterval(handle)

    Description: 
    Gets the interval setting of the given timer

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns the interval setting of the given timer
===============================================================================]]
function GetTimerInterval(handle)
	return handle:GetTimerInterval()
end

--[[=============================================================================
    c4_timer Unit Tests
===============================================================================]]
function __test_c4_timer()
	require "test.C4Virtual"
	require "lib.c4_log"
	require "common.c4_init"

	OnDriverInit()

	local LOG = c4_log:new("test_c4_timer")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	function OnTestTimerExpired()
		c4Timer:KillTimer()
	end

	-- create an instance of the timer
	c4Timer = c4_timer:new("Test", 45, "MINUTES", OnTestTimerExpired)

	assert(c4Timer._id == 0, "_id is not equal to '0' it is: " .. c4Timer._id)
	c4Timer:StartTimer()
	assert(c4Timer._id == 10001, "_id is not equal to '10001' it is: " .. c4Timer._id)
	assert(c4Timer:TimerStarted() == true, "TimerStarted is not equal to true it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == false, "TimerStopped is not equal to false it is: " .. tostring(c4Timer:TimerStopped()))
	OnTimerExpired(c4Timer._id)
	assert(c4Timer:TimerStarted() == false, "TimerStarted is not equal to false it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == true, "TimerStopped is not equal to true it is: " .. tostring(c4Timer:TimerStopped()))
end end)
package.preload['lib.c4_object'] = (function (...)
--[[=============================================================================
    c4_object Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_object = "2016.01.08"
end

function inheritsFrom( baseClass )
	local new_class = {}
	local class_mt = { __index = new_class }

	function new_class:create(...)
		local newinst = {}

		setmetatable( newinst, class_mt )

		-- Call the constructor when we create this class
		if newinst.construct then
			-- Allow returning a different obj than self. This allows for readonly tables etc...
			newinst = newinst:construct(...) or newinst
		end

		return newinst
	end

	if nil ~= baseClass then
		setmetatable( new_class, { __index = baseClass } )
	end

	--[[=============================================================================
	    Implementation of additional OO properties starts here 
	===============================================================================]]
	
	-- Return the class object of the instance
	function new_class:class()
		return new_class
	end

	--[[=============================================================================
	    Return the super class object of the instance. 

	    Note Calling methods on the base class itself will modify
	    the base table's static properties. In order to have call
	    the base class' methods and have them modify the current object
	    use super() or superAsSelf(). 
	===============================================================================]]
	function new_class:superClass()
		return baseClass
	end

	--[[=============================================================================
	    Returns a table that allows calling of the base class's method
	    while maintaining the objects state as the modified state of the base
	    class' methods. For example consider the following statements (order matters):

	    -- The child sees the parents property if the child hasn't overriden the property
	       obj:superClass().id = "parent"
	       obj.id == "parent" -- true

	    -- Setting the property on the child overrides (hides) the parents property
	       obj.id = "child"
	       obj.id == "child" -- true

	    -- The super() method pass
	       obj:super().id == "parent" -- true
	       obj:super().id = "child"
	       obj:super().id == "parent" -- still true
	       obj.id == "child" -- still true
	===============================================================================]]
	function new_class:super()
		local holder = {}

		holder.child = self
		holder.parent = baseClass

		local mt = {}
		mt.__index = function(table, index)
			if table.parent[index] then
				return table.parent[index]
			else 
				return table.child[index]
			end
		end

		-- Only set the new values to the child.
		mt.__newindex = function(table, key, value)
			table.child[key] = value
		end

		mt.__tostring = function(table)
			return tostring(table.child)
		end

		setmetatable(holder, mt)
		return holder
	end

	new_class.new = new_class.create

	--[[=============================================================================
	    Return true if the caller is an instance of theClass
	===============================================================================]]
	function new_class:isa( theClass )
		local b_isa = false
		local cur_class = new_class

		while ( nil ~= cur_class ) and ( false == b_isa ) do
			if cur_class == theClass then
				b_isa = true
			else
				cur_class = cur_class:superClass()
			end
		end

		return b_isa
	end

	return new_class
end

--[[=============================================================================
    Inheritance unit tests
===============================================================================]]
function __test_inheritance()
	local b = inheritsFrom(nil)

	b.construct = function(self, msg)
		self._msg = msg
	end

	local t = inheritsFrom(b)
	t.construct = function(self, msg)
		self:super():construct(msg)
	end

	t1 = t:new("t1")
	t2 = t:new("t2")
	assert(t1._msg == "t1", "t1 message is not equal to 't1' it''s: " .. t1._msg)
	assert(t2._msg == "t2", "t2 message is not equal to 't2' it''s: " .. t2._msg)    
	assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");
	assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

	t1:superClass().id = "parent"
	assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

	-- Setting the property on the child overrides (hides) the parents property
	t1.id = "child"
	assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

	-- The super() method maintains the self pointer to the child and not to the base
	assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
	t1:super().id = "child1"
	assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
	assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")
end end)
package.preload['lib.c4_log'] = (function (...)
--[[=============================================================================
    c4_log Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_log = "2016.01.08"
end

c4_log = inheritsFrom(nil)

function c4_log:construct(logName)
	self._logLevel = tonumber(string.sub(Properties['Log Level'] or "", 1, 1)) or 5
	self._outputPrint = Properties['Log Mode']:find("Print") ~= nil
	self._outputC4Log = Properties['Log Mode']:find("Log") ~= nil
	self._logName = logName or ""

	-- make sure Property is up to date (no harm if absent)
	C4:UpdateProperty("Log Level", Properties['Log Level'])
end

function c4_log:SetLogLevel(level)
	self._logLevel = tonumber(string.sub(level or "", 1, 1)) or self._logLevel
end

function c4_log:LogLevel()
	return self._logLevel
end

function c4_log:OutputPrint(value)
	self._outputPrint = value
end

function c4_log:OutputC4Log(value)
	self._outputC4Log = value
end

function c4_log:SetLogName(logName)

	if (logName == nil or logName == "") then
		logName = ""
	else
		logName = logName .. ": "
	end

	self._logName = logName
end

function c4_log:LogName()
	return self._logName
end

function c4_log:Enabled()
	return (self._outputPrint or self._outputC4Log)
end

function c4_log:PrintEnabled()
	return self._outputPrint
end

function c4_log:C4LogEnabled()
	return self._outputC4Log
end

function c4_log:CreateTableText(tValue, tableText)
	tableText = tableText or ""

	if (type(tValue) == "table") then

		tableText = tableText .. "{"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ","
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableText(v, tableText)
				tableText = tableText .. ","
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ","
			end
		end

		tableText = tableText .. "}"
	end

	return tableText
end

function InsertIndent(indentLevel)
	local indentStr = ""

	for i=1, indentLevel do
		indentStr = indentStr .. "\t"
	end

	return indentStr
end

function c4_log:CreateTableTextFormatted(tValue, tableText, indentLevel)
	tableText = tableText or ""
	indentLevel = indentLevel or 0

	if (type(tValue) == "table") then

		indentLevel = indentLevel + 1
		tableText = tableText .. "{\n"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. InsertIndent(indentLevel) .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. InsertIndent(indentLevel) .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ",\n"
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',\n"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableTextFormatted(v, tableText, indentLevel)
				tableText = tableText .. ",\n"
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ",\n"
			end
		end

		indentLevel = indentLevel - 1
		tableText = tableText .. InsertIndent(indentLevel) .. "}"
	end

	return tableText
end

MAX_TABLE_LEVELS = 10
function c4_log:PrintTable(tValue, tableText, sIndent, level)
	tableText = tableText or ""
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				if (tableText == "") then
					tableText = sIndent .. tostring(k) .. ":  " .. tostring(v)
					if (sIndent == ".   ") then sIndent = "    " end
				else
					tableText = tableText .. "\n" .. sIndent .. tostring(k) .. ":  " .. tostring(v)
				end
				if (type(v) == "table") then
					tableText = self:PrintTable(v, tableText, sIndent .. "   ", level)
				end
			end
		else
			tableText = tableText .. "\n" .. sIndent .. tostring(tValue)
		end
	end
	
	return tableText
end

function c4_log:LogTable(tValue, sIndent, level)
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				C4:ErrorLog(self._logName .. sIndent .. tostring(k) .. ":  " .. tostring(v))
				if (type(v) == "table") then
					self:LogTable(v, sIndent .. "   ", level)
				end
			end
		else
			C4:ErrorLog(self._logName .. sIndent .. tValue)
		end
	end
end

function c4_log:Print(logLevel, sLogText)

	if (self._logLevel >= logLevel) then
		if (type(sLogText) == "table") then
			if (self._outputPrint) then
				print (self:PrintTable(sLogText, tableText, ".   ", 0))
			end

			if (self._outputC4Log) then
				self:LogTable(sLogText, "   ", 0)
			end

			return
		end

		if (self._outputPrint) then
			print (sLogText)
		end

		if (self._outputC4Log) then
			C4:ErrorLog(self._logName .. tostring(sLogText))
		end
	end
end

function c4_log:Fatal(sLogText, ...)
	self:LogOutput(0, sLogText, ...)
end

function c4_log:Error(sLogText, ...)
	self:LogOutput(1, sLogText, ...)
end

function c4_log:Warn(sLogText, ...)
	self:LogOutput(2, sLogText, ...)
end

function c4_log:Info(sLogText, ...)
	self:LogOutput(3, sLogText, ...)
end

function c4_log:Debug(sLogText, ...)
	self:LogOutput(4, sLogText, ...)
end

function c4_log:Trace(sLogText, ...)
	self:LogOutput(5, sLogText, ...)
end

function c4_log:LogOutput(level, sLogText, ...)
	if (LogEnabled()) then
		if (type(sLogText) == "string") then
			sLogText = string.format(sLogText, ...)
		end

		self:Print(level, sLogText)
	end
end

--[[=============================================================================
    c4_log wrapper functions
===============================================================================]]
function TryLog(level, sLogText, ...)
	LOG:LogOutput(level, sLogText, ...)
end

--[[=============================================================================
    SetLogLevel(level)

    Description: 
    Sets the desired log level to view

    Parameters:
    level(int) - The logging level to set the message to
                 0 = Fatal
                 1 = Error
                 2 = Warn
                 3 = Info
                 4 = Debug
                 5 = Trace

    Returns:
    None
===============================================================================]]
function SetLogLevel(level)
	LOG:SetLogLevel(level)
end

--[[=============================================================================
    LogLevel()

    Description: 
    Returns the currently set log level

    Parameters:
    None

    Returns:
    The current log level
        0 = Fatal
        1 = Error
        2 = Warn
        3 = Info
        4 = Debug
        5 = Trace
===============================================================================]]
function LogLevel()
	return LOG:LogLevel()
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputPrint(value)
	LOG:OutputPrint(value)
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages to file or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputC4Log(value)
	LOG:OutputC4Log(value)
end

--[[=============================================================================
    SetLogName(logName)

    Description: 
    Sets the name of the log file where the messages will be written to

    Parameters:
    logName(string) - Sets the name of the log to write messages to

    Returns:
    None
===============================================================================]]
function SetLogName(logName)
	LOG:SetLogName(logName)
end

--[[=============================================================================
    LogName(logName)

    Description: 
    Gets the name of the log file where the messages will be written to

    Parameters:
    None

    Returns:
    The value of the log file that has been set
===============================================================================]]
function LogName()
	return LOG:LogName()
end

--[[=============================================================================
    LogEnabled()

    Description: 
    Identifies if logging or print has been enabled

    Parameters:
    None

    Returns:
    true if either logging or print has been enabled, false otherwise
===============================================================================]]
function LogEnabled()
	return LOG:Enabled()
end

--[[=============================================================================
    PrintEnabled()

    Description: 
    Gets the state of print output

    Parameters:
    None

    Returns:
    true if print has been enabled, false otherwise
===============================================================================]]
function PrintEnabled()
	return LOG:PrintEnabled()
end

--[[=============================================================================
    C4LogEnabled()

    Description: 
    Gets the state of logging

    Parameters:
    None

    Returns:
    true if logging has been enabled, false otherwise
===============================================================================]]
function C4LogEnabled()
	return LOG:C4LogEnabled()
end

--[[=============================================================================
    LogFatal(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Fatal(0) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogFatal(sLogText, ...)
	local status, err = pcall(TryLog, 0, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogFatal failed: " .. err)
	end
end

--[[=============================================================================
    LogError(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Error(1) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogError(sLogText, ...)
	local status, err = pcall(TryLog, 1, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogError failed: " .. err)
	end
end

--[[=============================================================================
    LogWarn(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Warn(2) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogWarn(sLogText, ...)
	local status, err = pcall(TryLog, 2, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogWarn failed: " .. err)
	end
end

--[[=============================================================================
    LogInfo(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Info(3) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogInfo(sLogText, ...)
	local status, err = pcall(TryLog, 3, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogInfo failed: " .. err)
	end
end

--[[=============================================================================
    LogDebug(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Debug(4) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogDebug(sLogText, ...)
	local status, err = pcall(TryLog, 4, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogDebug failed: " .. err)
	end
end

--[[=============================================================================
    LogTrace(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Trace(5) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogTrace(sLogText, ...)
	local status, err = pcall(TryLog, 5, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogTrace failed: " .. err)
	end
end

function dbgPrint(buf)
	if (LOG:PrintEnabled()) then
		print (buf)
	end
end

function dbgHexdump(buf)
	hexdump(buf, dbgPrint)
end

--[[=============================================================================
    c4_log unit tests
===============================================================================]]
function __test_c4_log()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	assert(LOG:LogName() == "test_c4_log", "_logName is not equal to 'test_c4_log' it is: " .. LOG:LogName())

	-- Test setting log level
	LOG:SetLogLevel("2 - Warning")
	assert(LOG:LogLevel() == 2, "_logLevel is not equal to '2' it is: " .. LOG:LogLevel())

	LOG:SetLogLevel(3)
	assert(LOG:LogLevel() == 3, "_logLevel is not equal to '3' it is: " .. LOG:LogLevel())

	-- Test enabling logs
	LOG:OutputPrint(false)
	assert(LOG:PrintEnabled() == false, "_outputPrint is not equal to 'false' it is: " .. tostring(LOG:PrintEnabled()))

	LOG:OutputC4Log(true)
	assert(LOG:C4LogEnabled() == true, "_outputC4Log is not equal to 'true' it is: " .. tostring(LOG:C4LogEnabled()))

	LOG:SetLogLevel(4)
	LogTrace("***** This is a test *****")
end

function __test_CreatTableText()
	local tTest = {}

	tTest[1] = {}
	tTest[2] = {}
	tTest[3] = 30
	tTest[4] = "Forty"
	
	LogTrace("----- tText -----")
	LogTrace(tTest)

	local tTest2 = { One = {},
					 Two = {},
					 Three = 30,
					 Four = "Forty" }
	LogTrace("----- tText2 -----")
	LogTrace(tTest2)
	
	local tTest3 = { [1] = {},
					 [2] = {},
					 [3] = 30,
					 [4] = "Forty" }
	LogTrace("----- tText3 -----")
	LogTrace(tTest3)

	local tTest4 = { [1] = {},
					 Two = {},
					 [3] = 30,
					 [4] = "Forty",
					 Five = "Fifty" }
	LogTrace("----- tText4 -----")
	LogTrace(tTest4)

	local tableText = LOG:CreateTableText(tTest4)
	LogTrace("----- tableText -----")
	LogTrace(tableText)
	
	--local tNew = {[1] = {},[3] = 30,[4] = 'Forty',Five = 'Fifty',Two = {},}
	--LogTrace(tNew)
end

function __TestCreateTableTextFormatted()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	local tButtons = {
		Name = 'heat',
		Attributes = {},
		ChildNodes = {
			[1] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '51',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Pool Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'POOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[2] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '53',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Spa Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'SPAHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[3] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '54',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {Value = 'Pool Solar Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {}
					},
					[3] = {
						Value = 'SOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				}
			}
		}
	}

	print(LOG:CreateTableTextFormatted(tButtons))
end end)
package.preload['common.c4_common'] = (function (...)
--[[=============================================================================
    ON_INIT, Timer,s and Property management functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_log"
require "lib.c4_timer"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_common = "2016.01.08"
end

--[[=============================================================================
    Create and Initialize Logging
===============================================================================]]
function ON_DRIVER_EARLY_INIT.c4_common()
	-- Create a logger
	LOG = c4_log:new("Template_c4z Change Name")
end

function ON_DRIVER_INIT.c4_common()
	-- Create Log Timer
	gC4LogTimer = c4_timer:new("Log Timer", 45, "MINUTES", OnLogTimerExpired)
end

--[[=============================================================================
    Log timer callback function
===============================================================================]]
function OnLogTimerExpired()
	LogWarn("Turning Log Mode Off (timer expired)")
	gC4LogTimer:KillTimer()
	
	C4:UpdateProperty("Log Mode", "Off")
	OnPropertyChanged("Log Mode")
end

gForceLogging = false
function ON_PROPERTY_CHANGED.LogMode(propertyValue)
	gC4LogTimer:KillTimer()
	
	if (gForceLogging) then
		LOG:OutputPrint(true)
		LOG:OutputC4Log(true)
	else
		LOG:OutputPrint(propertyValue:find("Print") ~= nil)
		LOG:OutputC4Log(propertyValue:find("Log") ~= nil)
		if (propertyValue == "Off") then
			return
		end
		
		gC4LogTimer:StartTimer()
	end
end

function ON_PROPERTY_CHANGED.LogLevel(propertyValue)
	if (gForceLogging) then
		LOG:SetLogLevel("5 - Debug")
	else
		LOG:SetLogLevel(propertyValue)
	end
end

--[[=============================================================================
    Print Template Versions
===============================================================================]]
function TemplateVersion()
	print ("\nTemplate Versions")
	print ("-----------------------")
	for k, v in pairs(TEMPLATE_VERSION) do
		print (k .. " = " .. v)
	end
	
	print ("")
end end)
package.preload['common.c4_driver_declarations'] = (function (...)
--[[=============================================================================
    Driver Declarations used to call startup routines, teardown routines, and 
    other basic functions of the drivers operation

    Copyright 2017 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Template Version Table
TEMPLATE_VERSION = {}
TEMPLATE_VERSION.c4_driver_declarations = "2017.04.25"

-- Command Handler Tables
EX_CMD = {}
PRX_CMD = {}
UI_REQ = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}
PROG_CONDITIONAL = {}


--[[=============================================================================
    Tables of functions
    The following tables are function containers that are called within the
    following functions:

    OnDriverInit()
        First calls all functions contained within ON_DRIVER_EARLY_INIT table
        then calls all functions contained within ON_DRIVER_INIT table

    OnDriverLateInit()
        Calls all functions contained within ON_DRIVER_LATEINIT table

    OnDriverDestroyed()
        Calls all functions contained within ON_DRIVER_DESTROYED table

    OnPropertyChanged()
        Calls all functions contained within ON_PROPERTY_CHANGED table
===============================================================================]]
ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001 end)
--[[=============================================================================
    Basic Template for TV Driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]------------
require "common.c4_driver_declarations"
require "common.c4_common"
require "common.c4_init"
require "common.c4_property"
require "common.c4_command"
require "common.c4_notify"
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_utils"
require "lib.c4_timer"
require "actions"
require "device_specific_commands"
require "device_messages"
require "projector_init"
require "properties"
require "proxy_commands"
require "connections"
require "projector.projector_proxy_class"
require "projector.projector_proxy_commands"
require "projector.projector_proxy_notifies"
require "minidriver"


-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.driver = "2015.03.31"
end

--[[=============================================================================
    Constants
===============================================================================]]
PROJECTOR_PROXY_BINDINGID = 5001
AMAZON_PROXY_BINDINGID = 5002
CHANNELPLUS_PROXY_BINDINGID = 5003
LGCONTENTSTORE_PROXY_BINDINGID = 5004
NETFLIX_PROXY_BINDINGID = 5005
VUDU_PROXY_BINDINGID = 5006
WEBBROWSER_PROXY_BINDINGID = 5007
YOUTUBE_PROXY_BINDINGID = 5008
GOOGLEPLAY_PROXY_BINDINGID = 5009
HULU_PROXY_BINDINGID = 5010
SLINGTV_PROXY_BINDINGID = 5011




VERSION = '3'


--[[=============================================================================
    Initialization Code
===============================================================================]]
function ON_DRIVER_EARLY_INIT.main()
	--UpdateProperty('Driver Version', VERSION)
end

function ON_DRIVER_INIT.main()
     DRIVER_NAME = C4:GetDriverConfigInfo("name")
	SetLogName(DRIVER_NAME)

	-- TODO: If cloud based driver then uncomment the following line
	--ConnectURL()
end

function ON_DRIVER_LATEINIT.main()
    --enable the Lua Command Window in Composer
    if (LuaC4Object.AllowExecute) then C4:AllowExecute(true) end   
    
end

function ON_DRIVER_EARLY_INIT.tv_driver()

end

function ON_DRIVER_INIT.tv_driver()
    -- TODO: Modify tVolumeRamping to have on entry per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {["state"] = false, ["mode"] = ""}
    gMuted = "" 	--valid values, "True" & "False"
    gInputSelection = {["InProgress"] = false, ["Command"] = ""}
    gCurrentlySelectedRFInputName = ""
    gDevMsgCallback = ""
    
    -- Create an instance of the TV Proxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local  bProcessesDeviceMessages = true
    gProjectorProxy = ProjectorProxy:new(PROJECTOR_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping)

    gPowerOnTimer = c4_timer:new("PowerOnTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnTimerExpired)
    gPowerOffTimer = c4_timer:new("PowerOffTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffTimerExpired)
    gSendPowerOnAfterPowerOffDelayTimer = c4_timer:new("SendPowerOnAfterPowerOffDelayTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnSendPowerOnAfterPowerOffDelayTimerExpired)
    gQueryPowerDelayTimer = c4_timer:new("QueryPowerDelayTimer", 1, "SECONDS", OnQueryPowerDelayTimerExpired)
    gSetInputDelayTimer = c4_timer:new("SetInputDelayTimer", 2, "SECONDS", OnSetInputDelayTimerExpired, command)
    gQueryInputDelayTimer = c4_timer:new("QueryInputDelayTimer", 2, "SECONDS", OnQueryInputDelayTimerExpired)
    gSetChannelDelayTimer = c4_timer:new("SetChannelDelayTimer", 2, "SECONDS", OnSetChannelDelayTimerExpired, command)
    gQueryChannelDelayTimer = c4_timer:new("QueryChannelDelayTimer", 2, "SECONDS", OnQueryChannelDelayTimerExpired)
    gQueryMuteDelayTimer = c4_timer:new("QueryMuteDelayTimer", 1, "SECONDS", OnQueryMuteDelayTimerExpired)
    gNetworkInitTimer = c4_timer:new("NetworkInitTimer", 5, "SECONDS", OnNetworkInitTimerExpired)
    gChannelBufferTimer = CreateTimer('ChannelBufferTimer', tonumber(Properties['Channel Buffer Timer Value']), 'SECONDS', 'sendBufferedChannel', false)
    gVolumeRampTimer = c4_timer:new("VolumeRampTimer", tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", OnVolumeRampTimerExpired)
    gMenuNavigationTimer = c4_timer:new("MenuNavigationTimer", tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", OnMenuNavigationTimerExpired)
    gTransportNavigationTimer = c4_timer:new("TransportNavigationTimer", tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", OnTransportNavigationTimerExpired)
    
    gWoLNetConnectDelayTimer = c4_timer:new("WoLNetConnectDelayTimer", 2, "SECONDS", OnWoLNetConnectDelayTimerExpired)
    
    gProxyNameToBindingIdMap = {}
    buildProxyNameToBindingIdMap()    

end

function ON_DRIVER_LATEINIT.tv_driver()
    DRIVER_CONTROL = C4:GetDriverConfigInfo("controlmethod") or ""
    if (DRIVER_CONTROL == "ir" or DRIVER_CONTROL == "serial") then
	   C4:SetPropertyAttribs ("Network Keep Alive Interval Seconds", 1)
	   C4:SetPropertyAttribs ("Password", 1)
	   C4:SetPropertyAttribs ("Network Connection", 1)
	   C4:SetPropertyAttribs ("Wired MAC Address", 1)
	   C4:SetPropertyAttribs ("Wi-Fi MAC Address", 1)
	   C4:SetPropertyAttribs ("Connected To Network", 1)
    end 
end

--[[=============================================================================
    Driver Code
===============================================================================]]

function buildProxyNameToBindingIdMap()
  local xml = C4:GetDriverConfigInfo("proxies")
  xml = "<proxies>" .. xml .. "</proxies>"
  local t = C4:ParseXml(xml)

  for j,k in pairs(t.ChildNodes) do 
    print(k.Attributes.name, k.Attributes.proxybindingid)
    gProxyNameToBindingIdMap[k.Attributes.name] = k.Attributes.proxybindingid
  end
end

function set_proxy_visbility(idBinding, value)
    local bHide = false
    if (value == "HIDE") then
	   bHide = true
	   SetProxyVisibilityInAllRooms(idBinding, bHide, "Hiding")
    elseif (value == "SHOW") then
	   bHide = false
	   SetProxyVisibilityInAllRooms(idBinding, bHide, "Showing")
    else
	   --do nothing
    end
end

function SetProxyVisibilityInAllRooms(idBinding, bHide, sShowHide)
  idBinding = idBinding or 0
  if (idBinding == 0) then return end -- silently fail if no binding passed in.

  -- Get Bound Proxy's Device ID / Name.
  local id, name = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), idBinding))

  LogTrace(sShowHide .. " '" .. name .. "' in all rooms...")

  -- Send hide command to all rooms, for 'ALL' Navigator groups.
  for roomid, roomname in pairs(C4:GetDevicesByC4iName("roomdevice.c4i") or {}) do
    LogTrace(sShowHide .. " '" .. name .. "' in '" .. roomname .. "' room.")
    C4:SendToDevice(roomid, "SET_DEVICE_HIDDEN_STATE", {PROXY_GROUP = "ALL", DEVICE_ID = id, IS_HIDDEN = bHide})
  end 
end

function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    --local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    --local delay_units = select(4, ...) or "MILLISECONDS"
    --LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end

	-- For LG command will contain SET_ID which must be replaced with hex string value of the id
	--strCommand=string.gsub(command, 'SET_ID', string.format ("%02X", Properties['Set ID']))

	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
	     local command_delay = 3
	     local delay_units = "SECONDS"
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
		
		cmd = lg_encrypt(cmd) 
		gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
    elseif (gControlMethod == "Serial") then
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
	     local command_delay = 3
	     local delay_units = "SECONDS"
		stx = ""
		etx = "\r"
		--cmd = stx .. command .. etx
		cmd = stx .. strCommand .. etx
		--using ACK, so we do not want to send delay values. The WaitResponseTimer (timout) will use the defualt value - 3 seconds
		--gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
		gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
    elseif (gControlMethod == "IR") then
	     local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
	     local delay_units = select(4, ...) or "MILLISECONDS"
		cmd = command
		gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    --gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
end

function PackAndQueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    --local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    --local delay_units = select(4, ...) or "MILLISECONDS"
    --LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    LogTrace("PackAndQueuePriority1Command(), command_name = " .. command_name)
    if (command == "") then
	   LogWarn("PackAndQueuePriority1Command(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueuePriority1Command()")
	   return
    end

	-- For LG command will contain SET_ID which must be replaced with hex string value of the id
	--strCommand=string.gsub(command, 'SET_ID', string.format ("%02X", Properties['Set ID']))

	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
	     local command_delay = 3
	     local delay_units = "SECONDS"
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
		
		cmd = lg_encrypt(cmd) 
		gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)
    elseif (gControlMethod == "Serial") then
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
	     local command_delay = 3
	     local delay_units = "SECONDS"
		stx = ""
		etx = "\r"
		--cmd = stx .. command .. etx
		cmd = stx .. strCommand .. etx
		--using ACK, so we do not want to send delay values. The WaitResponseTimer (timout) will use the defualt value - 3 seconds
		--gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
		gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)
    elseif (gControlMethod == "IR") then
	     local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
	     local delay_units = select(4, ...) or "MILLISECONDS"
		cmd = command
		gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)
    else
		LogWarn("PackAndQueuePriority1Command(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    --gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)
end

--[[=============================================================================
    PBKDF2 Encryption
===============================================================================]]
function ON_DRIVER_INIT.LG()
	--used by AES128 Encryption
	gKey = PBKDF2(Properties["Password"])
end

PBKDF2_DIGEST_TYPE = "SHA256"	-- get full list with C4:GetSupportedDigests()
PBKDF2_SALT = tohex("63 61 b8 0e 9b dc a6 63 8d 07 20 f2 cc 56 8f b9")
PBKDF2_ITERATIONS = 16384
PBKDF2_KEY_LENGTH = 16
PBKDF2_OPTIONS = {
  ["return_encoding"] = "NONE",	-- HEX, NONE, BASE64
  ["salt_encoding"] = "NONE",		-- HEX, NONE, BASE64
}

function PBKDF2(password)
    local pwd = C4:PBKDF2(PBKDF2_DIGEST_TYPE, password, PBKDF2_SALT, PBKDF2_ITERATIONS, PBKDF2_KEY_LENGTH, PBKDF2_OPTIONS)
    
    return pwd
end

--[[=============================================================================
    AES128 Encryption
===============================================================================]]

function lg_encrypt(command)
    print("lg_encrypt, command = " .. command)
    local IV = tohex("00000000000000000000000000000000") --will need to be generated randomly in next firmware...
    --local IV = get_random_key()
    local encrypt_cmd = Encrypt(IV, true) .. Encrypt(command, false)
    hexdump(encrypt_cmd)
    return encrypt_cmd
end

function get_random_key()
    local len = 16
    local iv = ""
	math.randomseed(os.time())
	for i=1, len do
		iv = iv .. string.char((math.floor(math.random() *100000000000)) % 255)
 	end
	return iv
end

function lg_decrypt(message)
     local length = string.len(message) 
     if (length < 32) then print("lg_decrypt: message length not valid") return end
	local iv_enc = string.sub(message, 1,16)
	local msg_enc = string.sub(message, 17, length)
	
	local iv = Decrypt(iv_enc)
	local msg = Decrypt(msg_enc, iv)
	
     local pattern = "(.+)\n()"
     local message, pos = string.match(msg, pattern)
	if (message == nil) then return "" end
    
	return message
end



--[[  Comments from partner-supplied C code...

This is an implementation of the AES128 algorithm, specifically ECB and CBC mode.

The implementation is verified against the test vectors in:
  National Institute of Standards and Technology Special Publication 800-38A 2001 ED

ECB-AES128
----------

  plain-text test vectors:
    6bc1bee22e409f96e93d7e117393172a
    ae2d8a571e03ac9c9eb76fac45af8e51
    30c81c46a35ce411e5fbc1191a0a52ef
    f69f2445df4f9b17ad2b417be66c3710

  key:
    2b7e151628aed2a6abf7158809cf4f3c

  resulting ciphers:
    3ad77bb40d7a3660a89ecaf32466ef97 
    f5d3d58503b9699de785895a96fdbaaf 
    43b1cd7f598ece23881b00e3ed030688 
    7b0c785e27e8ad3f8223207104725dd4 

NOTE: String length must be evenly divisible by 16byte (str_len % 16 == 0)
]]

CIPHER_TYPE = "AES-128-CBC"	
CIPHER_OPTIONS = {
  ["return_encoding"] = "NONE",
  ["key_encoding"] = "NONE",
  ["iv_encoding"] = "NONE",
  ["data_encoding"] = "NONE",
  ["padding"] = false,
}


-- Pad the input if not 16 byte divisible...
function Encrypt(PlainText, bIV)
  bIV = bIV or false
  local PaddedText = PlainText
  local PaddingLength = 16 - (#PlainText % 16)
  --if (PaddingLength == 16) then PaddingLength = 0 end
  if (bIV) then
    PaddingLength = 0
  else
    if (PaddingLength == 0) then PaddingLength = 16 end
  end
  print("Encrypt, PaddingLength = " .. PaddingLength)
  for pcount = 1, PaddingLength, 1 do
   PaddedText = PaddedText .. string.char(PaddingLength) --The value to be padded is the number of bytes to be padded. (not 0 as in previous spec)
  end
  
  return C4:Encrypt(CIPHER_TYPE, gKey, "", PaddedText, CIPHER_OPTIONS)
end

function Decrypt(CipherText, iv)
  iv = iv or ""
  return C4:Decrypt(CIPHER_TYPE, gKey, iv, CipherText, CIPHER_OPTIONS)
end

--moved to ON_DRIVER_INIT.main
--gKey = PBKDF2("ip-controlasekey")
--gKey = tohex("2b7e151628aed2a6abf7158809cf4f3c")


INPUT = {}
table.insert(INPUT, tohex("6bc1bee22e409f96e93d7e117393172a"))
table.insert(INPUT, tohex("ae2d8a571e03ac9c9eb76fac45af8e51"))
table.insert(INPUT, tohex("30c81c46a35ce411e5fbc1191a0a52ef"))
table.insert(INPUT, tohex("f69f2445df4f9b17ad2b417be66c3710"))


function TestInput(input)
  hexdump(input, function(s) print("Input: " .. s) end)
  local encrypted = Encrypt(input)
  hexdump(encrypted, function(s) print("Encrypted: " .. s) end)
  hexdump(Decrypt(encrypted), function(s) print("Decrypted: " .. s) end)
end

--[[
for k,v in pairs(INPUT) do
  TestInput(v)
end
--]]
 
